"""
App constants
"""

# Standard Library
import re

# Django
from django.utils.translation import gettext_lazy as _

# AA Intel Tool
import aa_intel_tool.parser.module.chatlist
import aa_intel_tool.parser.module.dscan
import aa_intel_tool.parser.module.fleetcomp

# All internal URLs need to start with this prefix
INTERNAL_URL_PREFIX = "-"


# Localised units
distance_units_on_grid: str = """
    km|m    # Client in: English, German, Chinese, French, Japanese, Korean, Spanish
    |км|м   # Russian
"""
distance_units_off_grid: str = """
    AU      # Client in: English, Chinese, Japanese, Korean, Spanish
    |UA     # Client in: French
    |AE     # German
    |а.е.   # Russian
"""

distance_units: str = f"{distance_units_on_grid}|{distance_units_off_grid}"


# Pre-compiled regex patterns used throughout the app
REGEX_PATTERN = {
    "chatlist": re.compile(pattern=r"(?im)^[a-zA-Z0-9\u0080-\uFFFF -_]{3,37}$"),
    "dscan": re.compile(
        pattern=rf"(?im)^(\d+)[\t](.*)[\t](.*)[\t](-|(.*) ({distance_units}))$",
        flags=re.VERBOSE,
    ),
    "fleetcomp": re.compile(
        pattern=r"(?im)^([a-zA-Z0-9 -_]{3,37})[\t](.*)[\t](.*)[\t](.*)[\t](.*)[\t]([0-5] - [0-5] - [0-5])([\t](.*))?$"  # pylint: disable=line-too-long
    ),
    "localised_on_grid": re.compile(
        pattern=rf"{distance_units_on_grid}", flags=re.VERBOSE
    ),
    "localised_off_grid": re.compile(
        pattern=rf"{distance_units_off_grid}", flags=re.VERBOSE
    ),
}

# Supported intel types and their parameters
SUPPORTED_INTEL_TYPES = {
    "chatlist": {
        "name": _("Chat list"),
        "parser": aa_intel_tool.parser.module.chatlist.parse,
        "pattern": REGEX_PATTERN["chatlist"],
        "template": "aa_intel_tool/views/scan/chatlist.html",
    },
    "dscan": {
        "name": _("D-Scan"),
        "parser": aa_intel_tool.parser.module.dscan.parse,
        "pattern": REGEX_PATTERN["dscan"],
        "template": "aa_intel_tool/views/scan/dscan.html",
    },
    "fleetcomp": {
        "name": _("Fleet composition"),
        "parser": aa_intel_tool.parser.module.fleetcomp.parse,
        "pattern": REGEX_PATTERN["fleetcomp"],
        "template": "aa_intel_tool/views/scan/fleetcomp.html",
    },
}



"""
Test cases for the chatlist parser module.
"""

# Standard Library
from unittest import TestCase
from unittest.mock import MagicMock, patch

# AA Intel Tool
from aa_intel_tool.exceptions import ParserError
from aa_intel_tool.models import Scan
from aa_intel_tool.parser.module.chatlist import parse


class TestParse(TestCase):
    """
    Test cases for the parse function in the chatlist module.
    """

    @patch(
        "aa_intel_tool.parser.module.chatlist.AppSettings.INTELTOOL_ENABLE_MODULE_CHATSCAN",
        True,
    )
    @patch("aa_intel_tool.parser.module.chatlist._get_character_info")
    @patch("aa_intel_tool.parser.module.chatlist._parse_chatscan_data")
    @patch("aa_intel_tool.parser.module.chatlist.safe_scan_to_db")
    def test_parses_valid_scan_data(
        self, mock_safe_scan_to_db, mock_parse_chatscan_data, mock_get_character_info
    ):
        """
        Test should parse valid scan data and return a Scan object.

        :param mock_safe_scan_to_db:
        :type mock_safe_scan_to_db:
        :param mock_parse_chatscan_data:
        :type mock_parse_chatscan_data:
        :param mock_get_character_info:
        :type mock_get_character_info:
        :return:
        :rtype:
        """

        mock_get_character_info.return_value = MagicMock()
        mock_parse_chatscan_data.return_value = {
            "pilots": [],
            "corporations": [],
            "alliances": [],
        }
        mock_safe_scan_to_db.return_value = Scan()

        scan_data = ["Character1", "Character2"]
        result = parse(scan_data)

        self.assertIsInstance(result, Scan)

    @patch(
        "aa_intel_tool.parser.module.chatlist.AppSettings.INTELTOOL_ENABLE_MODULE_CHATSCAN",
        True,
    )
    @patch("aa_intel_tool.parser.module.chatlist._get_character_info")
    @patch("aa_intel_tool.parser.module.chatlist._parse_chatscan_data")
    def test_returns_parsed_data_when_safe_to_db_is_false(
        self, mock_parse_chatscan_data, mock_get_character_info
    ):
        """
        Test should parse valid scan data and return a dictionary when safe_to_db is False.

        :param mock_parse_chatscan_data:
        :type mock_parse_chatscan_data:
        :param mock_get_character_info:
        :type mock_get_character_info:
        :return:
        :rtype:
        """

        mock_get_character_info.return_value = MagicMock()
        mock_parse_chatscan_data.return_value = {
            "pilots": [],
            "corporations": [],
            "alliances": [],
        }

        scan_data = ["Character1", "Character2"]
        result = parse(scan_data, safe_to_db=False)

        self.assertIsInstance(result, dict)

    @patch(
        "aa_intel_tool.parser.module.chatlist.AppSettings.INTELTOOL_ENABLE_MODULE_CHATSCAN",
        False,
    )
    def test_raises_error_when_module_disabled(self):
        """
        Test should raise a ParserError when the module is disabled.

        :return:
        :rtype:
        """

        with self.assertRaises(ParserError):
            parse([])

    @patch(
        "aa_intel_tool.parser.module.chatlist.AppSettings.INTELTOOL_ENABLE_MODULE_CHATSCAN",
        True,
    )
    @patch(
        "aa_intel_tool.parser.module.chatlist.AppSettings.INTELTOOL_CHATSCAN_MAX_PILOTS",
        1,
    )
    def test_raises_error_when_pilot_limit_exceeded(self):
        """
        Test should raise a ParserError when the pilot limit is exceeded.

        :return:
        :rtype:
        """

        scan_data = ["Character1", "Character2"]

        with self.assertRaises(ParserError):
            parse(scan_data)

    @patch(
        "aa_intel_tool.parser.module.chatlist.AppSettings.INTELTOOL_ENABLE_MODULE_CHATSCAN",
        True,
    )
    @patch("aa_intel_tool.parser.module.chatlist._get_character_info")
    @patch("aa_intel_tool.parser.module.chatlist._parse_chatscan_data")
    def test_handles_empty_scan_data(
        self, mock_parse_chatscan_data, mock_get_character_info
    ):
        """
        Test should handle empty scan data gracefully.

        :param mock_parse_chatscan_data:
        :type mock_parse_chatscan_data:
        :param mock_get_character_info:
        :type mock_get_character_info:
        :return:
        :rtype:
        """

        mock_get_character_info.return_value = MagicMock()
        mock_parse_chatscan_data.return_value = {
            "pilots": [],
            "corporations": [],
            "alliances": [],
        }

        scan_data = []
        result = parse(scan_data)

        self.assertIsInstance(result, Scan)



# Generated by Django 4.0.10 on 2023-09-22 21:21

# Django
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("aa_intel_tool", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="scan",
            name="raw_data",
            field=models.TextField(verbose_name="Scan raw data"),
        ),
    ]



"""
The views …
"""

# Django
from django.contrib import messages
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse
from django.shortcuts import redirect, render
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Intel Tool
from aa_intel_tool import __title__
from aa_intel_tool.app_settings import AppSettings
from aa_intel_tool.constants import SUPPORTED_INTEL_TYPES
from aa_intel_tool.exceptions import ParserError
from aa_intel_tool.form import IntelForm
from aa_intel_tool.models import Scan, ScanData
from aa_intel_tool.parser.general import parse_intel

logger = LoggerAddTag(my_logger=get_extension_logger(name=__name__), prefix=__title__)


def index(request: WSGIRequest) -> HttpResponse:
    """
    Intel tool index (basically just the form)

    :return:
    :rtype:
    """

    if request.method == "POST":
        # Create a form instance and populate it with data from the request
        form = IntelForm(data=request.POST)

        # Check whether it's valid:
        if form.is_valid():
            exception_caught = False
            scan_data = form.cleaned_data["eve_intel"]

            try:
                parsed_intel = parse_intel(form_data=scan_data)

            # Catching our own parser exceptions
            except ParserError as exc:
                exception_caught = True
                errormessage = _("The provided data could not be parsed.") + f" ({exc})"
                messages.error(request=request, message=errormessage)

            # Catching every other exception we can't think of (hopefully)
            except Exception as exc:  # pylint: disable=broad-exception-caught
                exception_caught = True
                errormessage = (
                    _("(System Error) Something unexpected happened.") + f" ({exc})"
                )
                messages.error(request=request, message=errormessage)

            if exception_caught:
                return redirect(to="aa_intel_tool:intel_tool_index")

            return redirect(to="aa_intel_tool:intel_tool_scan", scan_hash=parsed_intel)

        context = {"form": form, "app_settings": AppSettings}

    # If a GET (or any other method) we'll create a blank form
    else:
        form = IntelForm()
        context = {"form": form, "app_settings": AppSettings}

    return render(
        request=request,
        template_name="aa_intel_tool/views/index.html",
        context=context,
    )


def scan(request: WSGIRequest, scan_hash: str):
    """
    Scan view

    :param request:
    :type request:
    :param scan_hash:
    :type scan_hash:
    :return:
    :rtype:
    """

    try:
        # pylint: disable=no-member
        intel_scan = Scan.objects.exclude(scan_type=Scan.Type.INVALID).get(pk=scan_hash)
    except Scan.DoesNotExist:
        messages.error(
            request=request,
            message=_("The scan you were looking for could not be found."),
        )

        return redirect(to="aa_intel_tool:intel_tool_index")

    logger.debug(msg=f"Intel Type: {intel_scan.scan_type}")

    scan_data = {
        "scan_type": intel_scan.scan_type,
        "created": intel_scan.created,
        "raw_data": intel_scan.raw_data,
    }

    context = {
        "scan_hash": scan_hash,
        "scan": scan_data,
        "scan_data_section": ScanData.Section,
        "parser_title": SUPPORTED_INTEL_TYPES[intel_scan.scan_type]["name"],
        "app_settings": AppSettings,
    }

    if intel_scan.scan_type in SUPPORTED_INTEL_TYPES:
        return render(
            request=request,
            template_name=SUPPORTED_INTEL_TYPES[intel_scan.scan_type]["template"],
            context=context,
        )

    messages.error(
        request=request,
        message=_("The scan you were looking for could not be found."),
    )

    return redirect(to="aa_intel_tool:intel_tool_index")



"""
Test for admin.py
"""

# Django
from django.contrib import admin
from django.test import RequestFactory, TestCase
from django.urls import reverse

# AA Intel Tool
from aa_intel_tool.admin import BaseReadOnlyAdminMixin, ScanAdmin
from aa_intel_tool.models import Scan


class TestBaseReadOnlyAdminMixin(TestCase):
    """
    Test the BaseReadOnlyAdminMixin class
    """

    def test_add_permission_is_denied(self):
        """
        Test if admin.BaseReadOnlyAdminMixin.has_add_permission returns False

        :return:
        :rtype:
        """

        request = RequestFactory().get(
            path=reverse(viewname="aa_intel_tool:intel_tool_index")
        )

        self.assertFalse(BaseReadOnlyAdminMixin.has_add_permission(request))

    def test_change_permission_is_denied(self):
        """
        Test if admin.BaseReadOnlyAdminMixin.has_change_permission returns False

        :return:
        :rtype:
        """

        request = RequestFactory().get(
            path=reverse(viewname="aa_intel_tool:intel_tool_index")
        )

        self.assertFalse(BaseReadOnlyAdminMixin.has_change_permission(request))

    def test_delete_permission_is_denied(self):
        """
        Test if admin.BaseReadOnlyAdminMixin.has_delete_permission returns False

        :return:
        :rtype:
        """

        request = RequestFactory().get(
            path=reverse(viewname="aa_intel_tool:intel_tool_index")
        )

        self.assertFalse(BaseReadOnlyAdminMixin.has_delete_permission(request))


class TestScanAdmin(TestCase):
    """
    Test the ScanAdmin class
    """

    def test_displays_scan_type_with_link(self):
        """
        Test if ScanAdmin._scan_type displays the scan type with a link

        :return:
        :rtype:
        """

        scan = Scan(raw_data="test data")
        scan.save()

        admin_instance = ScanAdmin(model=Scan, admin_site=admin.site)
        result = admin_instance._scan_type(scan)

        self.assertIn("Open in a new browser tab", result)
        self.assertIn(scan.hash, result)

    def test_displays_raw_data_in_pre_tag(self):
        """
        Test if ScanAdmin._raw_data displays the raw data in a <pre> tag

        :return:
        :rtype:
        """

        scan = Scan.objects.create(raw_data="test data")

        admin_instance = ScanAdmin(model=Scan, admin_site=admin.site)
        result = admin_instance._raw_data(scan)

        self.assertIn("<pre>test data</pre>", result)



"""
Test the app settings in local.py
"""

# Standard Library
from unittest import mock

# Django
from django.conf import settings
from django.test import TestCase, override_settings

# AA Intel Tool
from aa_intel_tool.app_settings import AppSettings, debug_enabled

SETTINGS_PATH = "aa_intel_tool.app_settings"


class TestAppSettings(TestCase):
    """
    Tests for App Settings
    """

    def test_scan_retention_time_default(self):
        """
        Test for the default INTELTOOL_SCAN_RETENTION_TIME

        :return:
        :rtype:
        """

        retention_time = AppSettings.INTELTOOL_SCAN_RETENTION_TIME
        expected_retention_time = 30

        self.assertEqual(first=retention_time, second=expected_retention_time)

    @mock.patch(SETTINGS_PATH + ".AppSettings.INTELTOOL_SCAN_RETENTION_TIME", 75)
    def test_scan_retention_time_custom(self):
        """
        Test for a custom INTELTOOL_SCAN_RETENTION_TIME

        :return:
        :rtype:
        """

        retention_time = AppSettings.INTELTOOL_SCAN_RETENTION_TIME
        expected_retention_time = 75

        self.assertEqual(first=retention_time, second=expected_retention_time)

    @override_settings()
    def test_chatscan_max_pilots_default(self):
        """
        Test for the default INTELTOOL_CHATSCAN_MAX_PILOTS

        :return:
        :rtype:
        """

        del settings.INTELTOOL_CHATSCAN_MAX_PILOTS

        max_pilots = AppSettings.INTELTOOL_CHATSCAN_MAX_PILOTS
        expected_max_pilots = 500

        print("max_pilots:", max_pilots)

        self.assertEqual(first=max_pilots, second=expected_max_pilots)

    @mock.patch(SETTINGS_PATH + ".AppSettings.INTELTOOL_CHATSCAN_MAX_PILOTS", 1000)
    def test_chatscan_max_pilots_custom(self):
        """
        Test for a custom INTELTOOL_CHATSCAN_MAX_PILOTS

        :return:
        :rtype:
        """

        max_pilots = AppSettings.INTELTOOL_CHATSCAN_MAX_PILOTS
        expected_max_pilots = 1000

        self.assertEqual(first=max_pilots, second=expected_max_pilots)

    def test_dscan_grid_size(self):
        """
        Test for the default INTELTOOL_DSCAN_GRID_SIZE

        :return:
        :rtype:
        """

        grid_size = AppSettings.INTELTOOL_DSCAN_GRID_SIZE
        expected_grid_size = 10000

        self.assertEqual(first=grid_size, second=expected_grid_size)

    @mock.patch(SETTINGS_PATH + ".AppSettings.INTELTOOL_DSCAN_GRID_SIZE", 1000)
    def test_dscan_grid_size_custom(self):
        """
        Test for a custom INTELTOOL_DSCAN_GRID_SIZE

        :return:
        :rtype:
        """

        grid_size = AppSettings.INTELTOOL_DSCAN_GRID_SIZE
        expected_grid_size = 1000

        self.assertEqual(first=grid_size, second=expected_grid_size)

    @override_settings(DEBUG=True)
    def test_debug_enabled_with_debug_true(self) -> None:
        """
        Test debug_enabled with DEBUG = True

        :return:
        :rtype:
        """

        self.assertTrue(debug_enabled())

    @override_settings(DEBUG=False)
    def test_debug_enabled_with_debug_false(self) -> None:
        """
        Test debug_enabled with DEBUG = False

        :return:
        :rtype:
        """

        self.assertFalse(debug_enabled())



"""
The models
"""

# Django
from django.db import models, transaction
from django.utils.crypto import get_random_string
from django.utils.translation import gettext_lazy as _


class Scan(models.Model):
    """
    The scans
    """

    class Type(models.TextChoices):
        """
        The choices for Scan.type
        """

        INVALID = "invalid", _("Invalid scan data")
        DSCAN = "dscan", _("D-Scan")
        FLEETCOMP = "fleetcomp", _("Fleet composition")
        CHATLIST = "chatlist", _("Chat list")

    hash = models.CharField(
        primary_key=True,
        editable=False,
        unique=True,
        max_length=30,
        verbose_name=_("Scan hash"),
    )

    created = models.DateTimeField(
        auto_now_add=True,
        db_index=True,
        verbose_name=_("Creation date/time"),
    )

    raw_data = models.TextField(verbose_name=_("Scan raw data"))

    scan_type = models.CharField(
        max_length=9,
        choices=Type.choices,
        default=Type.INVALID,
        verbose_name=_("Scan type"),
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        default_permissions = ()
        verbose_name = _("Scan")
        verbose_name_plural = _("Scans")

    def __str__(self) -> str:
        return str(self.pk)

    @transaction.atomic()
    def save(self, *args, **kwargs):
        """
        Generates the scan hash on save

        :param args:
        :type args:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """

        if self._state.adding is True:
            self.hash = Scan.generate_scan_hash()

        super().save(*args, **kwargs)

        if self.hash == "":
            self.hash = Scan.generate_scan_hash()
            self.save()

    @staticmethod
    def generate_scan_hash():
        """
        Get a random string we can use as hash

        :return:
        :rtype:
        """

        scan_hash = get_random_string(length=30)

        while Scan.objects.filter(hash=scan_hash).exists():  # pylint: disable=no-member
            scan_hash = get_random_string(length=30)

        return scan_hash


class ScanData(models.Model):
    """
    Scan section data
    """

    class Section(models.TextChoices):
        """
        The choices for ScanData.section
        """

        INVALID = "invalid", _("Invalid scan data")
        PILOTLIST = "pilotlist", _("Pilot list")
        CORPORATIONLIST = "corporationlist", _("Corporation list")
        ALLIANCELIST = "alliancelist", _("Alliance list")
        SHIPTYPES = "shiptypes", _("Ship types")
        SHIPLIST = "shiplist", _("Ship list")
        SHIPLIST_ON_GRID = "shiplist_on_grid", _("Ship list (on grid)")
        SHIPLIST_OFF_GRID = "shiplist_off_grid", _("Ship list (off grid)")
        STRUCTURES_ON_GRID = "structures_on_grid", _("Structures (on grid)")
        STARBASES_ON_GRID = "starbases_on_grid", _("Starbases (on grid)")
        DEPLOYABLES_ON_GRID = "deployables_on_grid", _("Deployables (on grid)")
        MISCELLANEOUS_ON_GRID = "miscellaneous_on_grid", _("Miscellaneous (on grid)")
        SOLAR_SYSTEM_INFORMATION = "solar_system_information", _("System information")
        FLEETCOMPOSITION = "fleetcomposition", _("Fleet composition")

    scan = models.ForeignKey(
        Scan,
        related_name="scan_data",
        null=True,
        blank=True,
        default=None,
        on_delete=models.CASCADE,
        verbose_name=_("Scan"),
    )

    section = models.CharField(
        max_length=24,
        choices=Section.choices,
        default=Section.INVALID,
        verbose_name=_("Scan section"),
    )

    processed_data = models.JSONField(
        default=dict, verbose_name=_("Processed scan data")
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        default_permissions = ()
        verbose_name = _("Scan data")
        verbose_name_plural = _("Scan data")

        unique_together = ("scan", "section")



"""
Helper for our tests
"""

# Standard Library
from pathlib import Path

# Django
from django.core.handlers.wsgi import WSGIRequest
from django.template import Context, Template


def render_template(string, context=None):
    """
    Helper to render templates
    :param string:
    :param context:
    :return:
    """

    context = context or {}
    context = Context(dict_=context)

    return Template(template_string=string).render(context=context)


def load_chatscan_txt() -> str:
    """
    Loading chatscan.txt

    :return:
    :rtype:
    """

    return (Path(__file__).parent / "test-data/chatscan.txt").read_text()


def load_chatscan_faulty_txt() -> str:
    """
    Loading chatscan-faulty.txt

    :return:
    :rtype:
    """

    return (Path(__file__).parent / "test-data/chatscan-faulty.txt").read_text()


def load_dscan_txt() -> str:
    """
    Loading dscan.txt

    :return:
    :rtype:
    """

    return (Path(__file__).parent / "test-data/dscan.txt").read_text()


def load_fleetcomp_txt() -> str:
    """
    Loading fleetcomp.txt

    :return:
    :rtype:
    """

    return (Path(__file__).parent / "test-data/fleetcomp.txt").read_text()


def response_content_to_str(response: WSGIRequest) -> str:
    """
    Return the content of a WSGIRequest response as string

    :param response:
    :type response:
    :return:
    :rtype:
    """

    return response.content.decode(response.charset)



"""
Test for the data_structures helper => aa_intel_tool/helper/data_structure.py
"""

# Django
from django.test import TestCase

# AA Intel Tool
from aa_intel_tool.helper.data_structure import dict_to_list


class TestHelperDataStructures(TestCase):
    """
    Test the data_structures helper
    """

    def test_dict_to_list(self):
        """
        Test that we get a list from a dict

        :return:
        :rtype:
        """

        my_dict = {
            "foobar": {"foo": "bar", "bar": "foo"},
            "barfoo": {"bar": "foo", "foo": "bar"},
        }
        my_list = dict_to_list(my_dict)
        expected_list = [{"bar": "foo", "foo": "bar"}, {"foo": "bar", "bar": "foo"}]

        self.assertListEqual(list1=my_list, list2=expected_list)



# Generated by Django 4.0.10 on 2023-08-28 22:17

# Django
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Scan",
            fields=[
                (
                    "hash",
                    models.CharField(
                        editable=False,
                        max_length=30,
                        primary_key=True,
                        serialize=False,
                        unique=True,
                        verbose_name="Scan hash",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        auto_now_add=True,
                        db_index=True,
                        verbose_name="Creation date/time",
                    ),
                ),
                ("raw_data", models.TextField(verbose_name="Scan RAW data")),
                (
                    "scan_type",
                    models.CharField(
                        choices=[
                            ("invalid", "Invalid scan data"),
                            ("dscan", "D-Scan"),
                            ("fleetcomp", "Fleet composition"),
                            ("chatlist", "Chat list"),
                        ],
                        default="invalid",
                        max_length=9,
                        verbose_name="Scan type",
                    ),
                ),
            ],
            options={
                "verbose_name": "Scan",
                "verbose_name_plural": "Scans",
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="ScanData",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "section",
                    models.CharField(
                        choices=[
                            ("invalid", "Invalid scan data"),
                            ("pilotlist", "Pilot list"),
                            ("corporationlist", "Corporation list"),
                            ("alliancelist", "Alliance list"),
                            ("shiptypes", "Ship types"),
                            ("shiplist", "Ship list"),
                            ("shiplist_on_grid", "Ship list (on grid)"),
                            ("shiplist_off_grid", "Ship list (off grid)"),
                            ("structures_on_grid", "Structures (on grid)"),
                            ("starbases_on_grid", "Starbases (on grid)"),
                            ("deployables_on_grid", "Deployables (on grid)"),
                            ("miscellaneous_on_grid", "Miscellaneous (on grid)"),
                            ("solar_system_information", "System information"),
                            ("fleetcomposition", "Fleet composition"),
                        ],
                        default="invalid",
                        max_length=24,
                        verbose_name="Scan section",
                    ),
                ),
                (
                    "processed_data",
                    models.JSONField(default=dict, verbose_name="Processed scan data"),
                ),
                (
                    "scan",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="scan_data",
                        to="aa_intel_tool.scan",
                        verbose_name="Scan",
                    ),
                ),
            ],
            options={
                "verbose_name": "Scan data",
                "verbose_name_plural": "Scan data",
                "default_permissions": (),
                "unique_together": {("scan", "section")},
            },
        ),
    ]



"""
Tests for the helper functions in the eve_character module.
"""

# Standard Library
from unittest.mock import Mock, patch

# Django
from django.test import TestCase

# Alliance Auth
from allianceauth.eveonline.models import (
    EveAllianceInfo,
    EveCharacter,
    EveCorporationInfo,
)

# AA Intel Tool
from aa_intel_tool.helper.eve_character import (
    _create_alliance,
    _create_character,
    _create_corporation,
    get_or_create_character,
)


class TestCreateAlliance(TestCase):
    """
    Test the _create_alliance function.
    """

    def test_creates_new_alliances(self):
        """
        Test that new alliances are created when they do not exist in the database.

        :return:
        :rtype:
        """

        alliance_ids = [1001, 1002, 1003]

        with (
            patch.object(
                EveAllianceInfo.objects,
                "filter",
                return_value=EveAllianceInfo.objects.none(),
            ),
            patch.object(
                EveAllianceInfo.objects, "create_alliance"
            ) as mock_create_alliance,
        ):
            _create_alliance(alliance_ids)

            self.assertEqual(mock_create_alliance.call_count, 3)

    def test_does_not_create_existing_alliances(self):
        """
        Test that existing alliances are not created again.

        :return:
        :rtype:
        """

        alliance_ids = [1001, 1002, 1003]

        # Create existing alliances in the database
        EveAllianceInfo.objects.create(
            alliance_id=1001, alliance_name="Alliance 1001", executor_corp_id=2001
        )
        EveAllianceInfo.objects.create(
            alliance_id=1002, alliance_name="Alliance 1002", executor_corp_id=2002
        )

        with (
            patch.object(
                EveAllianceInfo.objects,
                "filter",
                wraps=EveAllianceInfo.objects.filter,
            ),
            patch.object(
                EveAllianceInfo.objects, "create_alliance"
            ) as mock_create_alliance,
        ):
            _create_alliance(alliance_ids)

            mock_create_alliance.assert_called_once_with(alliance_id=1003)

    def test_handles_empty_alliance_ids(self):
        """
        Test that no alliances are created when the input list is empty.

        :return:
        :rtype:
        """

        alliance_ids = []

        with (
            patch.object(EveAllianceInfo.objects, "filter"),
            patch.object(
                EveAllianceInfo.objects, "create_alliance"
            ) as mock_create_alliance,
        ):
            _create_alliance(alliance_ids)

            mock_create_alliance.assert_not_called()


class TestCreateCorporation(TestCase):
    """
    Test the _create_corporation function.
    """

    def test_creates_new_corporations(self):
        """
        Test that new corporations are created when they do not exist in the database.

        :return:
        :rtype:
        """

        corporation_ids = [2001, 2002, 2003]

        with (
            patch.object(
                EveCorporationInfo.objects,
                "filter",
                return_value=EveCorporationInfo.objects.none(),
            ),
            patch.object(
                EveCorporationInfo.objects, "create_corporation"
            ) as mock_create_corporation,
        ):
            _create_corporation(corporation_ids)

            self.assertEqual(mock_create_corporation.call_count, 3)

    def test_does_not_create_existing_corporations(self):
        """
        Test that existing corporations are not created again.

        :return:
        :rtype:
        """

        corporation_ids = [2001, 2002, 2003]

        EveCorporationInfo.objects.create(
            corporation_id=2001, corporation_name="Corporation 2001", member_count=10
        )
        EveCorporationInfo.objects.create(
            corporation_id=2002, corporation_name="Corporation 2002", member_count=20
        )

        with (
            patch.object(
                EveCorporationInfo.objects,
                "filter",
                wraps=EveCorporationInfo.objects.filter,
            ),
            patch.object(
                EveCorporationInfo.objects, "create_corporation"
            ) as mock_create_corporation,
        ):
            _create_corporation(corporation_ids)

            mock_create_corporation.assert_called_once_with(corp_id=2003)

    def test_handles_empty_corporation_ids(self):
        """
        Test that no corporations are created when the input list is empty.

        :return:
        :rtype:
        """

        corporation_ids = []

        with (
            patch.object(EveCorporationInfo.objects, "filter"),
            patch.object(
                EveCorporationInfo.objects, "create_corporation"
            ) as mock_create_corporation,
        ):
            _create_corporation(corporation_ids)

            mock_create_corporation.assert_not_called()


class TestCreateCharacter(TestCase):
    """
    Test the _create_character function.
    """

    @patch("aa_intel_tool.helper.eve_character.EveCharacter.objects.create_character")
    @patch("aa_intel_tool.helper.eve_character.EveAllianceInfo.objects.create_alliance")
    @patch("allianceauth.eveonline.providers.provider.get_corp")
    def test_creates_new_characters(
        self, mock_get_corp, mock_create_alliance, mock_create_character
    ):
        # Setup mock return values
        mock_create_character.side_effect = [
            Mock(character_id=1001, alliance_id=3001, corporation_id=2001),
            Mock(character_id=1002, alliance_id=3001, corporation_id=2002),
            Mock(character_id=1003, alliance_id=None, corporation_id=2003),
        ]

        mock_get_corp_side_effect = [
            Mock(
                id=2001,
                name="Corporation 2001",
                ticker="TICK1",
                alliance_id=3001,
                members=10,
            ),
            Mock(
                id=2002,
                name="Corporation 2002",
                ticker="TICK2",
                alliance_id=3001,
                members=20,
            ),
            Mock(
                id=2003,
                name="Corporation 2003",
                ticker="TICK3",
                alliance_id=None,
                members=30,
            ),
        ]

        # Ensure the mock objects return plain values for attributes
        for mock_corp in mock_get_corp_side_effect:
            mock_corp.name = "Corporation " + str(mock_corp.id)
            mock_corp.ticker = "TICK" + str(mock_corp.id % 1000)
            mock_corp.members = mock_corp.members

        mock_get_corp.side_effect = (
            lambda *args, **kwargs: mock_get_corp_side_effect.pop(0)
        )

        # Call the function under test
        _create_character(character_ids=[1, 2, 3], with_affiliation=True)

        # Assert that create_character was called three times
        self.assertEqual(mock_create_character.call_count, 3)

        # Assert that create_alliance was called once with the correct ID
        mock_create_alliance.assert_called_once_with(alliance_id=3001)

    def test_handles_empty_character_ids(self):
        """
        Test that no characters are created when the input list is empty.

        :return:
        :rtype:
        """

        character_ids = []

        with (
            patch.object(EveCharacter.objects, "filter"),
            patch.object(
                EveCharacter.objects, "create_character"
            ) as mock_create_character,
        ):
            _create_character(character_ids)

            mock_create_character.assert_not_called()

    @patch("aa_intel_tool.helper.eve_character.EveCharacter.objects.create_character")
    @patch("aa_intel_tool.helper.eve_character.EveAllianceInfo.objects.create_alliance")
    @patch("allianceauth.eveonline.providers.provider.get_corp")
    def creates_characters_without_affiliation(
        self, mock_get_corp, mock_create_alliance, mock_create_character
    ):
        # Setup mock return values
        mock_create_character.side_effect = [
            Mock(character_id=1001, alliance_id=3001, corporation_id=2001),
            Mock(character_id=1002, alliance_id=3001, corporation_id=2002),
            Mock(character_id=1003, alliance_id=None, corporation_id=2003),
        ]

        mock_get_corp_side_effect = [
            Mock(
                id=2001,
                name="Corporation 2001",
                ticker="TICK1",
                alliance_id=3001,
                members=10,
            ),
            Mock(
                id=2002,
                name="Corporation 2002",
                ticker="TICK2",
                alliance_id=3001,
                members=20,
            ),
            Mock(
                id=2003,
                name="Corporation 2003",
                ticker="TICK3",
                alliance_id=None,
                members=30,
            ),
        ]

        # Ensure the mock objects return plain values for attributes
        for mock_corp in mock_get_corp_side_effect:
            mock_corp.name = "Corporation " + str(mock_corp.id)
            mock_corp.ticker = "TICK" + str(mock_corp.id % 1000)
            mock_corp.members = mock_corp.members

        mock_get_corp.side_effect = (
            lambda *args, **kwargs: mock_get_corp_side_effect.pop(0)
        )

        # Call the function under test
        _create_character(character_ids=[1, 2, 3], with_affiliation=False)

        # Assert that create_character was called three times
        self.assertEqual(mock_create_character.call_count, 3)

        # Assert that create_alliance was called once with the correct ID
        mock_create_alliance.assert_called_once_with(alliance_id=3001)


class TestGetOrCreateCharacter(TestCase):
    """
    Test the get_or_create_character function.
    """

    @patch("aa_intel_tool.helper.eve_character._create_character")
    def test_returns_existing_characters(self, mock_create_character):
        """
        Test that existing characters are returned without creating new ones.

        :param mock_create_character:
        :type mock_create_character:
        :return:
        :rtype:
        """

        character_ids = [3001, 3002]
        EveCharacter.objects.create(
            character_id=3001,
            character_name="Character 3001",
            corporation_id=2001,
            alliance_id=1001,
        )
        EveCharacter.objects.create(
            character_id=3002,
            character_name="Character 3002",
            corporation_id=2002,
            alliance_id=1002,
        )

        characters = get_or_create_character(character_ids)

        self.assertEqual(characters.count(), 2)
        mock_create_character.assert_not_called()

    @patch("aa_intel_tool.helper.eve_character._create_character")
    def test_creates_missing_characters(self, mock_create_character):
        """
        Test that missing characters are created when they do not exist in the database.

        :param mock_create_character:
        :type mock_create_character:
        :return:
        :rtype:
        """

        character_ids = [3001, 3002, 3003]
        EveCharacter.objects.create(
            character_id=3001,
            character_name="Character 3001",
            corporation_id=2001,
            alliance_id=1001,
        )

        characters = get_or_create_character(character_ids)

        self.assertEqual(characters.count(), 1)
        mock_create_character.assert_called_once_with(
            character_ids={3002, 3003}, with_affiliation=True
        )

    @patch("aa_intel_tool.helper.eve_character._create_character")
    def handles_empty_character_ids(self, mock_create_character):
        character_ids = []

        characters = get_or_create_character(character_ids)

        self.assertEqual(characters.count(), 0)
        mock_create_character.assert_not_called()

    @patch("aa_intel_tool.helper.eve_character._create_character")
    def handles_none_character_ids(self, mock_create_character):
        character_ids = None

        characters = get_or_create_character(character_ids)

        self.assertEqual(characters.count(), 0)
        mock_create_character.assert_not_called()



"""
Django admin integration
"""

# Django
from django.contrib import admin
from django.utils import html, safestring
from django.utils.translation import gettext_lazy as _

# Alliance Auth (External Libs)
from app_utils.urls import reverse_absolute

# AA Intel Tool
from aa_intel_tool.models import Scan


class BaseReadOnlyAdminMixin:
    """
    Base "Read Only" mixin for admin models
    """

    actions = None  # Removes the default delete action.

    @staticmethod
    def has_add_permission(request):  # pylint: disable=unused-argument
        """
        Has "add" permissions

        :param request:
        :type request:
        :return:
        :rtype:
        """

        return False

    @staticmethod
    def has_change_permission(request, obj=None):  # pylint: disable=unused-argument
        """
        Has "change" permissions

        :param request:
        :type request:
        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        return False

    @staticmethod
    def has_delete_permission(request, obj=None):  # pylint: disable=unused-argument
        """
        Has "delete" permissions

        :param request:
        :type request:
        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        return False


@admin.register(Scan)
class ScanAdmin(BaseReadOnlyAdminMixin, admin.ModelAdmin):
    """
    Scan Admin
    """

    list_display = ("hash", "scan_type", "created")
    fields = ("_scan_type", "_raw_data")

    ordering = ("-created",)

    @admin.display(description=_("Scan type"))
    def _scan_type(self, obj) -> str:
        """
        Add link to open the scan in a new browser tab

        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        intel_type = obj.get_scan_type_display()
        scan_link = reverse_absolute(
            viewname="aa_intel_tool:intel_tool_scan", args=[obj.hash]
        )
        link_text = _("Open in a new browser tab")

        return safestring.mark_safe(
            f'{intel_type} (<a href="{scan_link}" target="_blank" rel="noreferer noopener">{link_text}</a>)'
        )

    @admin.display(description=_("Raw data"))
    def _raw_data(self, obj) -> str:
        """
        Format the output properly

        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        return html.format_html("<pre>{}</pre>", obj.raw_data)



"""
Ajax views
"""

# Django
from django.core.handlers.wsgi import WSGIRequest
from django.http import JsonResponse

# AA Intel Tool
from aa_intel_tool.models import ScanData


def get_scan_data(
    request: WSGIRequest,  # pylint: disable=unused-argument
    scan_hash: str,
    scan_section: ScanData.Section,
) -> JsonResponse:
    """
    Get scan data for a specific section

    :param request:
    :type request:
    :param scan_hash:
    :type scan_hash:
    :param scan_section:
    :type scan_section:
    :return:
    :rtype:
    """

    try:
        scan_data = ScanData.objects.filter(  # pylint: disable=no-member
            scan_id__exact=scan_hash, section__exact=scan_section
        ).get()
        processed_data = scan_data.processed_data
    except ScanData.DoesNotExist:  # pylint: disable=no-member
        processed_data = {}

    return JsonResponse(data=processed_data, safe=False)



"""
Tests for the parsers' DB helper => aa_intel_tool/parser/helper/db.py
"""

# Django
from django.test import TestCase

# AA Intel Tool
from aa_intel_tool.models import Scan, ScanData
from aa_intel_tool.parser.helper.db import safe_scan_to_db


class TestParserHelperDb(TestCase):
    """
    The tests
    """

    def test_safe_scan_to_db_without_associated_data(self):
        """
        Test that the scan is successfully saved to the DB

        :return:
        :rtype:
        """

        new_scan = safe_scan_to_db(scan_type=Scan.Type.DSCAN, parsed_data={})

        from_db = Scan.objects.get(pk=new_scan.hash, scan_type__exact=Scan.Type.DSCAN)

        self.assertEqual(first=new_scan.hash, second=from_db.hash)

    def test_safe_scan_to_db_with_associated_data(self):
        """
        Test that the scan is successfully saved to the DB

        :return:
        :rtype:
        """

        parsed_data = {
            "foobar": {
                "section": ScanData.Section.PILOTLIST,
                "data": {"name": "William Riker"},
            }
        }

        new_scan = safe_scan_to_db(scan_type=Scan.Type.DSCAN, parsed_data=parsed_data)

        scan_from_db = Scan.objects.get(
            pk=new_scan.hash, scan_type__exact=Scan.Type.DSCAN
        )
        scan_data_from_db = ScanData.objects.get(
            scan_id__exact=scan_from_db.hash,
            section__exact=ScanData.Section.PILOTLIST,
        )

        self.assertEqual(first=new_scan.hash, second=scan_from_db.hash)
        self.assertEqual(
            first=scan_data_from_db.processed_data, second=parsed_data["foobar"]["data"]
        )



"""
Helper for different data structures
"""


def dict_to_list(input_dict: dict) -> list:
    """
    Cleaning up and sorting a dict and transform it into a list

    :param input_dict:
    :type input_dict:
    :return:
    :rtype:
    """

    return list(dict(sorted(input_dict.items())).values())



"""
App settings
"""

# Standard Library
from enum import IntEnum
from re import RegexFlag

# Django
from django.conf import settings

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.app_settings import clean_setting
from app_utils.logging import LoggerAddTag

# AA Intel Tool
from aa_intel_tool import __title__

logger = LoggerAddTag(my_logger=get_extension_logger(name=__name__), prefix=__title__)


class AppSettings:
    """
    App settings from local.py
    """

    # Enable or disable the chat scan module
    # Disabled by default
    INTELTOOL_ENABLE_MODULE_CHATSCAN = clean_setting(
        name="INTELTOOL_ENABLE_MODULE_CHATSCAN", default_value=False, required_type=bool
    )

    # Enable or disable the d-scan module
    # Enabled by default
    INTELTOOL_ENABLE_MODULE_DSCAN = clean_setting(
        name="INTELTOOL_ENABLE_MODULE_DSCAN", default_value=True, required_type=bool
    )

    # Enable or disable the fleet composition module
    # Enabled by default
    INTELTOOL_ENABLE_MODULE_FLEETCOMP = clean_setting(
        name="INTELTOOL_ENABLE_MODULE_FLEETCOMP",
        default_value=True,
        required_type=bool,
    )

    # Scan retention time
    # Sets the time in days for how long the scans will be kept in the database.
    # Set to 0 to keep scans indefinitely.
    INTELTOOL_SCAN_RETENTION_TIME = clean_setting(
        name="INTELTOOL_SCAN_RETENTION_TIME", default_value=30, required_type=int
    )

    # Set the maximum number of pilots allowed per chat scan
    # Set to 0 for no limit.
    INTELTOOL_CHATSCAN_MAX_PILOTS = clean_setting(
        name="INTELTOOL_CHATSCAN_MAX_PILOTS", default_value=500, required_type=int
    )

    # Set the grid size for D-Scans.
    # This defines the size of the grid in which ships and
    # structure are considered to be "on grid"
    INTELTOOL_DSCAN_GRID_SIZE = clean_setting(
        name="INTELTOOL_DSCAN_GRID_SIZE", default_value=10000, required_type=int
    )


class AdditionalEveCategoryId(IntEnum):
    """
    Eve category IDs which are not covered by Eve Universe
    Unfortunately Python doesn't allow to extend eveuniverse.constants.EveCategoryId
    """

    DEPLOYABLE = 22
    STARBASE = 23
    SCANNER_PROBE = 479


class UpwellStructureId(IntEnum):
    """
    Upwell Structure IDs
    """

    ANSIBLEX_JUMP_GATE = 35841


def debug_enabled() -> RegexFlag:
    """
    Check if DEBUG is enabled

    :return:
    :rtype:
    """

    return settings.DEBUG



"""
Chat list parser
"""

# Standard Library
from collections import defaultdict

# Django
from django.db.models import QuerySet
from django.utils.translation import gettext_lazy as _
from django.utils.translation import ngettext

# Alliance Auth
from allianceauth.eveonline.evelinks import dotlan, eveimageserver, evewho, zkillboard
from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag
from eveuniverse.models import EveEntity

# AA Intel Tool
from aa_intel_tool import __title__
from aa_intel_tool.app_settings import AppSettings
from aa_intel_tool.exceptions import ParserError
from aa_intel_tool.helper.data_structure import dict_to_list
from aa_intel_tool.helper.eve_character import get_or_create_character
from aa_intel_tool.models import Scan, ScanData
from aa_intel_tool.parser.helper.db import safe_scan_to_db

logger = LoggerAddTag(my_logger=get_extension_logger(name=__name__), prefix=__title__)


def _get_character_info(scan_data: list) -> QuerySet[EveCharacter]:
    """
    Get Eve character information and affiliation from a list of character names

    :param scan_data:
    :type scan_data:
    :return:
    :rtype:
    """

    # Excluding corporation_id=1000001 (Doomheim) to potentially force an update here …
    eve_characters = EveCharacter.objects.filter(character_name__in=scan_data).exclude(
        corporation_id=1000001
    )

    # Check if we have to bother Eve Universe or if we have all characters already
    if len(scan_data) != eve_characters.count():
        try:
            eve_character_ids = (
                EveEntity.objects.fetch_by_names_esi(names=scan_data, update=True)
                .filter(category=EveEntity.CATEGORY_CHARACTER)
                .values_list("id", flat=True)
            )
        except EveEntity.DoesNotExist as exc:  # pylint: disable=no-member
            raise ParserError(
                message=str(
                    _(
                        "Something went wrong while fetching the character information from ESI."
                    )
                )
            ) from exc

        # In case the name does not belong to an Eve character,
        # EveEntity returns an empty object
        if not eve_character_ids:
            raise ParserError(message=str(_("Character unknown to ESI.")))

        eve_characters = get_or_create_character(character_ids=eve_character_ids)

    return eve_characters


def _get_unaffiliated_alliance_info() -> dict:
    """
    Get the alliance_info dict for characters that are in no alliance

    :return:
    :rtype:
    """

    return {
        "id": 1,
        "name": "",
        "ticker": "",
        "logo": eveimageserver.alliance_logo_url(alliance_id=1, size=32),
    }


def _parse_alliance_info(
    eve_character: EveCharacter, with_evelinks: bool = True
) -> dict:
    """
    Parse the alliance information from an EveCharacter

    :param eve_character:
    :type eve_character:
    :return:
    :rtype:
    """

    if eve_character.alliance_id is None:
        return _get_unaffiliated_alliance_info()

    alliance_info = {
        "id": eve_character.alliance_id,
        "name": eve_character.alliance_name,
        "ticker": eve_character.alliance_ticker,
        "logo": eve_character.alliance_logo_url_32,
    }

    # Add eve links if requested
    if with_evelinks:
        alliance_info.update(
            {
                "dotlan": dotlan.alliance_url(eve_character.alliance_name),
                "zkillboard": zkillboard.alliance_url(eve_character.alliance_id),
            }
        )

    return alliance_info


def _parse_corporation_info(
    eve_character: EveCharacter,
    with_alliance_info: bool = True,
    with_evelinks: bool = True,
) -> dict:
    """
    Parse the corporation information from an EveCharacter

    :param eve_character:
    :type eve_character:
    :return:
    :rtype:
    """

    corporation_info = {
        "id": eve_character.corporation_id,
        "name": eve_character.corporation_name,
        "ticker": eve_character.corporation_ticker,
        "logo": eve_character.corporation_logo_url_32,
    }

    # Add eve links if requested
    if with_evelinks:
        corporation_info.update(
            {
                "dotlan": dotlan.corporation_url(name=eve_character.corporation_name),
                "zkillboard": zkillboard.corporation_url(
                    eve_id=eve_character.corporation_id
                ),
            }
        )

    # Add alliance info if requested
    if with_alliance_info:
        corporation_info["alliance"] = _parse_alliance_info(
            eve_character=eve_character, with_evelinks=with_evelinks
        )

    return corporation_info


def _parse_character_info(eve_character: EveCharacter) -> dict:
    """
    Parse the character information from an EveCharacter

    :param eve_character:
    :type eve_character:
    :return:
    :rtype:
    """

    return {
        "id": eve_character.character_id,
        "name": eve_character.character_name,
        "portrait": eve_character.portrait_url_32,
        "evewho": evewho.character_url(eve_character.character_id),
        "zkillboard": zkillboard.character_url(eve_character.character_id),
        "corporation": _parse_corporation_info(
            eve_character=eve_character, with_alliance_info=False, with_evelinks=False
        ),
        "alliance": _parse_alliance_info(
            eve_character=eve_character, with_evelinks=False
        ),
    }


def _parse_chatscan_data(eve_characters: QuerySet[EveCharacter]) -> dict:
    """
    Parse the chat scan data and return character information,
    corporation information and alliance information for each character

    :param eve_characters:
    :type eve_characters:
    :return:
    :rtype:
    """

    counter = defaultdict(int)
    alliance_info = {}
    corporation_info = {}
    character_info = {}

    # Loop through the characters
    for eve_character in list(eve_characters):
        alliance_name = eve_character.alliance_name or "Unaffiliated"
        corporation_name = eve_character.corporation_name

        counter[alliance_name] += 1
        counter[corporation_name] += 1

        # Alliance Info
        if alliance_name not in alliance_info:
            alliance_info[alliance_name] = _parse_alliance_info(eve_character)

        # Corporation Info
        if corporation_name not in corporation_info:
            corporation_info[corporation_name] = _parse_corporation_info(eve_character)

        # Character Info
        character_info[eve_character.character_name] = _parse_character_info(
            eve_character
        )

        # Update the counter
        alliance_info[alliance_name]["count"] = counter[alliance_name]
        corporation_info[corporation_name]["count"] = counter[corporation_name]

    return {
        "pilots": dict_to_list(input_dict=character_info),
        "corporations": dict_to_list(input_dict=corporation_info),
        "alliances": dict_to_list(input_dict=alliance_info),
    }


def parse(
    scan_data: list, safe_to_db: bool = True, ignore_limit: bool = False
) -> Scan | dict:
    """
    Parse chat list

    :param scan_data:
    :type scan_data:
    :param safe_to_db:
    :type safe_to_db:
    :param ignore_limit:
    :type ignore_limit:
    :return:
    :rtype:
    """

    # Only parse the chat scan if the module is enabled
    if not AppSettings.INTELTOOL_ENABLE_MODULE_CHATSCAN:
        raise ParserError(message=str(_("The chat list module is currently disabled.")))

    logger.debug(msg=f"{len(scan_data)} name(s) to work through …")

    pilots_in_scan = len(scan_data)
    max_allowed_pilots = AppSettings.INTELTOOL_CHATSCAN_MAX_PILOTS

    # Check if the number of pilots in the scan exceeds the maximum allowed number
    if 0 < max_allowed_pilots < pilots_in_scan and not ignore_limit:
        logger.debug(
            msg=(
                f"Number of pilots in scan ({pilots_in_scan}) exceeds the maximum "
                f"allowed number ({max_allowed_pilots}). Throwing a tantrum …"
            )
        )

        # Throw a tantrum
        raise ParserError(
            message=ngettext(
                singular=f"Chat scans are currently limited to a maximum of {max_allowed_pilots} pilot per scan. Your list of pilots exceeds this limit.",
                plural=f"Chat scans are currently limited to a maximum of {max_allowed_pilots} pilots per scan. Your list of pilots exceeds this limit.",
                number=max_allowed_pilots,
            )
        )

    eve_characters = _get_character_info(scan_data=scan_data)
    logger.debug(msg=f"Got {len(eve_characters)} EveCharacter object(s) back from AA …")

    # Parse the data
    parsed_chatscan = _parse_chatscan_data(eve_characters=eve_characters)
    parsed_data = {
        "pilots": {
            "section": ScanData.Section.PILOTLIST,
            "data": parsed_chatscan["pilots"],
        },
        "corporations": {
            "section": ScanData.Section.CORPORATIONLIST,
            "data": parsed_chatscan["corporations"],
        },
        "alliances": {
            "section": ScanData.Section.ALLIANCELIST,
            "data": parsed_chatscan["alliances"],
        },
    }

    return (
        parsed_data
        if not safe_to_db
        else safe_scan_to_db(scan_type=Scan.Type.CHATLIST, parsed_data=parsed_data)
    )



"""
Hook into AA
"""

# Alliance Auth
from allianceauth import hooks
from allianceauth.services.hooks import MenuItemHook, UrlHook

# AA Intel Tool
from aa_intel_tool import __title__, urls


class AaIntelToolMenuItem(MenuItemHook):  # pylint: disable=too-few-public-methods
    """
    This class ensures only authorized users will see the menu entry
    """

    def __init__(self):
        # setup menu entry for sidebar
        MenuItemHook.__init__(
            self,
            text=__title__,
            classes="fa-solid fa-clipboard-list",
            url_name="aa_intel_tool:intel_tool_index",
            navactive=["aa_intel_tool:"],
        )

    def render(self, request):
        """
        Render the menu item

        :param request:
        :type request:
        :return:
        :rtype:
        """

        return MenuItemHook.render(self, request=request)


@hooks.register(name="menu_item_hook")
def register_menu():
    """
    Register our menu item

    :return:
    :rtype:
    """

    return AaIntelToolMenuItem()


@hooks.register(name="url_hook")
def register_urls():
    """
    Register our base url

    :return:
    :rtype:
    """

    return UrlHook(
        urls=urls,
        namespace="aa_intel_tool",
        base_url=r"^intel/",
        excluded_views=[
            "aa_intel_tool.views.ajax.get_scan_data",
            "aa_intel_tool.views.general.index",
            "aa_intel_tool.views.general.scan",
        ],
    )



"""
Fleet composition parser
"""

# Standard Library
import re

# Django
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.eveonline.evelinks import eveimageserver, evewho, zkillboard
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag
from eveuniverse.models import EveEntity, EveType

# AA Intel Tool
from aa_intel_tool import __title__
from aa_intel_tool.app_settings import AppSettings
from aa_intel_tool.exceptions import ParserError
from aa_intel_tool.helper.data_structure import dict_to_list
from aa_intel_tool.models import Scan, ScanData
from aa_intel_tool.parser.helper.db import safe_scan_to_db
from aa_intel_tool.parser.module.chatlist import _get_character_info
from aa_intel_tool.parser.module.chatlist import parse as parse_pilots

logger = LoggerAddTag(my_logger=get_extension_logger(name=__name__), prefix=__title__)


def get_fleet_composition(pilots: dict, ships: dict) -> dict:
    """
    Getting the fleet composition

    :param ships:
    :type ships:
    :return:
    :rtype:
    """

    # Get ship class IDs
    ship_class_ids = (
        EveEntity.objects.fetch_by_names_esi(names=ships["class"], update=True)
        .filter(category=EveEntity.CATEGORY_INVENTORY_TYPE)
        .values_list("id", flat=True)
    )

    # Get ship class details
    ship_class_details = EveType.objects.bulk_get_or_create_esi(
        ids=set(ship_class_ids), include_children=True
    ).values_list("id", "name", "eve_group__id", "eve_group__name", "mass", named=True)

    # Build ship class and type dictionaries
    for ship_class in list(ship_class_details):
        # Build ship class dict
        ships["class"][ship_class.name].update(
            {
                "id": ship_class.id,
                "name": ship_class.name,
                "type_id": ship_class.eve_group__id,
                "type_name": ship_class.eve_group__name,
                "image": eveimageserver.type_icon_url(type_id=ship_class.id, size=32),
                "mass": ship_class.mass * ships["class"][ship_class.name]["count"],
            }
        )

        # Build ship type dict
        ships["type"][ship_class.eve_group__name].update(
            {
                "id": ship_class.eve_group__id,
                "name": ship_class.eve_group__name,
            }
        )

    # Pilots
    pilot_details = _get_character_info(scan_data=list(pilots))

    # Build pilots dictionary
    for pilot in list(pilot_details):
        pilot_ship_class = next(
            ship_class
            for ship_class in ship_class_details
            if ship_class.name == pilots[pilot.character_name]["ship"]
        )

        pilots[pilot.character_name].update(
            {
                "id": pilot.character_id,
                "portrait": pilot.portrait_url_32,
                "evewho": evewho.character_url(pilot.character_id),
                "zkillboard": zkillboard.character_url(pilot.character_id),
                "ship_id": pilot_ship_class.id,
                "ship_type_id": pilot_ship_class.eve_group__id,
            }
        )

    return {
        "classes": dict_to_list(input_dict=ships["class"]),
        "types": dict_to_list(input_dict=ships["type"]),
        "pilots": dict_to_list(input_dict=pilots),
    }


def parse_line(line) -> list:
    """
    Parse a line from the fleet composition scan

    :param line:
    :type line:
    :return:
    :rtype:
    """

    # Let's split this list up
    #
    # line[0] => Pilot Name
    # line[1] => System
    # line[2] => Ship Class
    # line[3] => Ship Type
    # line[4] => Position in Fleet
    # line[5] => Skills (FC - WC - SC)
    # line[6] => Wing Name / Squad Name
    line = re.split(r"\t+", line.rstrip("\t"))

    return line if len(line) > 6 else line + [""]


def update_ships(ships, line) -> dict:
    """
    Update the ships dict

    :param ships:
    :type ships:
    :param line:
    :type line:
    :return:
    :rtype:
    """

    ships["class"].setdefault(line[2], {"count": 0})["count"] += 1
    ships["type"].setdefault(line[3], {"count": 0})["count"] += 1

    return ships


def handle_fleet_composition_and_participation(pilots, ships) -> tuple:
    """
    Handle the fleet composition and participation

    :param pilots:
    :type pilots:
    :param ships:
    :type ships:
    :return:
    :rtype:
    """

    fleet_composition = get_fleet_composition(pilots=pilots, ships=ships)
    participation = (
        parse_pilots(scan_data=list(set(pilots)), safe_to_db=False, ignore_limit=True)
        if AppSettings.INTELTOOL_ENABLE_MODULE_CHATSCAN
        else None
    )

    return fleet_composition, participation


def parse(scan_data: list) -> Scan:
    """
    Parse the fleet composition scan

    :param scan_data:
    :type scan_data:
    :return:
    :rtype:
    """

    if not AppSettings.INTELTOOL_ENABLE_MODULE_FLEETCOMP:
        raise ParserError(
            message=str(_("The fleet composition module is currently disabled."))
        )

    parsed_data = {}
    pilots = {}
    ships = {"type": {}, "class": {}}

    # Loop through the scan data
    for entry in scan_data:
        line = parse_line(entry)

        pilots[line[0]] = {
            "name": line[0],
            "solarsystem": line[1],
            "ship": line[2],
            "ship_type": line[3],
        }
        ships = update_ships(ships=ships, line=line)

    fleet_composition, participation = handle_fleet_composition_and_participation(
        pilots=pilots, ships=ships
    )

    logger.debug(msg=fleet_composition)

    # Add parsed data when available
    if fleet_composition["classes"]:
        parsed_data["classes"] = {
            "section": ScanData.Section.SHIPLIST,
            "data": fleet_composition["classes"],
        }
    if fleet_composition["types"]:
        parsed_data["shiptypes"] = {
            "section": ScanData.Section.SHIPTYPES,
            "data": fleet_composition["types"],
        }
    if fleet_composition["pilots"]:
        parsed_data["pilots_flying"] = {
            "section": ScanData.Section.FLEETCOMPOSITION,
            "data": fleet_composition["pilots"],
        }
    if participation:
        parsed_data.update(participation)

    return safe_scan_to_db(scan_type=Scan.Type.FLEETCOMP, parsed_data=parsed_data)



"""
AA Intel Tool exceptions
"""

# Django
from django.utils.translation import gettext_lazy as _


class AaIntelToolException(Exception):
    """
    An exception generated by AA Intel Tool.
    """


class ParserError(AaIntelToolException):
    """
    Parser Error raised for errors within our parsers
    """

    def __init__(self, message: str) -> None:
        super().__init__(message)

        self._message = message

    @property
    def message(self) -> str:
        """
        Message of the error.

        :return:
        :rtype:
        """

        return self._message

    def __str__(self) -> str:
        """
        String representation of the error.

        :return:
        :rtype:
        """

        message = self.message if self.message else _("Unknown parser error")

        return str(_(f"A parser error occurred » {message}"))



"""
D-Scan parser
"""

# Standard Library
import re
from collections import defaultdict

# Django
from django.db.models import QuerySet
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.eveonline.evelinks import eveimageserver
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag
from eveuniverse.constants import EveCategoryId
from eveuniverse.models import EveType

# AA Intel Tool
from aa_intel_tool import __title__
from aa_intel_tool.app_settings import (
    AdditionalEveCategoryId,
    AppSettings,
    UpwellStructureId,
)
from aa_intel_tool.exceptions import ParserError
from aa_intel_tool.helper.data_structure import dict_to_list
from aa_intel_tool.models import Scan, ScanData
from aa_intel_tool.parser.helper.db import safe_scan_to_db

logger = LoggerAddTag(my_logger=get_extension_logger(name=__name__), prefix=__title__)


def _is_on_grid(distance: str) -> bool:
    """
    Determine if something is "on grid" or not

    :param distance:
    :type distance:
    :return:
    :rtype:
    """

    # AA Intel Tool
    from aa_intel_tool.constants import (  # pylint: disable=import-outside-toplevel
        REGEX_PATTERN,
    )

    # Check if we have a distance
    match = re.search(pattern=REGEX_PATTERN["localised_on_grid"], string=distance)

    # Check if the distance is within the grid size
    if match:
        distance_sanitised = int(re.sub(pattern=r"[^0-9]", repl="", string=distance))
        return distance_sanitised <= AppSettings.INTELTOOL_DSCAN_GRID_SIZE

    return False


def _get_type_info_dict(eve_type: tuple) -> dict:
    """
    Get the eve_type info dict

    eve_type[0] = ID
    eve_type[1] = Name
    eve_type[2] = Group ID
    eve_type[3] = Group Name

    :param eve_type:
    :type eve_type:
    :return:
    :rtype:
    """

    return {
        "id": eve_type[0],
        "name": eve_type[1],
        "type_id": eve_type[2],
        "type_name": eve_type[3],
        "image": eveimageserver.type_icon_url(type_id=eve_type[0], size=32),
    }


def _get_ships(eve_types: QuerySet, counter: dict) -> dict:
    """
    Get the ships
    This will be the content of the following tables in the D-Scan view:
    » All Ships
    » On Grid
    » Off Grid
    » Ship Types

    :param eve_types:
    :type eve_types:
    :param counter:
    :type counter:
    :return:
    :rtype:
    """

    ships = {"all": {}, "ongrid": {}, "offgrid": {}, "types": {}}

    def _add_ship_info(ship_dict: dict, eve_type: tuple, count: int):
        """
        Add ship info to the ship_dict

        :param ship_dict:
        :type ship_dict:
        :param eve_type:
        :type eve_type:
        :param count:
        :type count:
        :return:
        :rtype:
        """

        if eve_type[1] not in ship_dict:
            ship_dict[eve_type[1]] = _get_type_info_dict(eve_type=eve_type)
            ship_dict[eve_type[1]]["count"] = count
            ship_dict[eve_type[1]]["mass"] = eve_type[4] * count

    # Ship types
    # Get all ship types
    #
    # Ship type list:
    #   - eve_type[0] = ID
    #   - eve_type[1] = Name
    #   - eve_type[2] = Group ID
    #   - eve_type[3] = Group Name
    #   - eve_type[4] = Mass
    eve_types_ships = eve_types.filter(
        eve_group__eve_category_id__exact=EveCategoryId.SHIP
    )

    # Loop through all ships types
    for eve_type in list(eve_types_ships):
        # Info for "All Ships" table
        if eve_type[0] in counter["all"]:
            _add_ship_info(
                ship_dict=ships["all"],
                eve_type=eve_type,
                count=counter["all"][eve_type[0]],
            )

        # Info for "On Grid" table
        if eve_type[0] in counter["ongrid"]:
            _add_ship_info(
                ship_dict=ships["ongrid"],
                eve_type=eve_type,
                count=counter["ongrid"][eve_type[0]],
            )

        # Info for "Off Grid" table
        if eve_type[0] in counter["offgrid"]:
            _add_ship_info(
                ship_dict=ships["offgrid"],
                eve_type=eve_type,
                count=counter["offgrid"][eve_type[0]],
            )

        # Info for "Ship Types" table
        if eve_type[3] not in ships["types"]:
            ships["types"][eve_type[3]] = {
                "id": eve_type[2],
                "name": eve_type[3],
                "count": 0,
            }

        # Add the count to the ship types
        ships["types"][eve_type[3]]["count"] += counter["all"][eve_type[0]]

    return {
        "all": dict_to_list(input_dict=ships["all"]),
        "ongrid": dict_to_list(input_dict=ships["ongrid"]),
        "offgrid": dict_to_list(input_dict=ships["offgrid"]),
        "types": dict_to_list(input_dict=ships["types"]),
    }


def _get_upwell_structures_on_grid(
    eve_types: QuerySet, counter: dict, ansiblex_destination: str = None
) -> list:
    """
    Get all Upwell structures that are on grid

    :param eve_types:
    :type eve_types:
    :param counter:
    :type counter:
    :return:
    :rtype:
    """

    eve_types_structures = eve_types.filter(
        eve_group__eve_category_id__exact=EveCategoryId.STRUCTURE
    )

    upwell_structures = {
        eve_type[1]: {
            **_get_type_info_dict(eve_type=eve_type),
            "count": counter["ongrid"][eve_type[0]],
            "name": (
                f"{eve_type[1]} » {ansiblex_destination}"
                if eve_type[0] == UpwellStructureId.ANSIBLEX_JUMP_GATE
                and ansiblex_destination
                else eve_type[1]
            ),
        }
        for eve_type in list(eve_types_structures)
        if eve_type[0] in counter["ongrid"]
    }

    return dict_to_list(input_dict=upwell_structures)


def _get_deployables_on_grid(eve_types: QuerySet, counter: dict) -> list:
    """
    Get all deployables that are on grid

    :param eve_types:
    :type eve_types:
    :param counter:
    :type counter:
    :return:
    :rtype:
    """

    eve_types_deployables = eve_types.filter(
        eve_group__eve_category_id__exact=AdditionalEveCategoryId.DEPLOYABLE
    )

    deployables = {
        eve_type[1]: {
            **_get_type_info_dict(eve_type=eve_type),
            "count": counter["ongrid"][eve_type[0]],
        }
        for eve_type in list(eve_types_deployables)
        if eve_type[0] in counter["ongrid"]
    }

    return dict_to_list(input_dict=deployables)


def _get_starbases_on_grid(eve_types: QuerySet, counter: dict) -> list:
    """
    Get all starbases and starbase modules that are on grid

    :param eve_types:
    :type eve_types:
    :param counter:
    :type counter:
    :return:
    :rtype:
    """

    eve_types_starbase = eve_types.filter(
        eve_group__eve_category_id__exact=AdditionalEveCategoryId.STARBASE
    )

    starbases = {
        eve_type[1]: {
            **_get_type_info_dict(eve_type=eve_type),
            "count": counter["ongrid"][eve_type[0]],
        }
        for eve_type in list(eve_types_starbase)
        if eve_type[0] in counter["ongrid"]
    }

    return dict_to_list(input_dict=starbases)


def _get_ansiblex_jumpgate_destination(ansiblex_name: str) -> str:
    """
    Get the Ansiblex Jump Gate destination system

    :param ansiblex_name:
    :type ansiblex_name:
    :return:
    :rtype:
    """

    return re.split(
        pattern=r" » ", string=re.split(pattern=r" - ", string=ansiblex_name)[0]
    )[1]


def _get_scan_details(scan_data: list) -> tuple:
    """
    Split the D-Scan data into more convenient parts

    :param scan_data:
    :type scan_data:
    :return:
    :rtype:
    """

    # AA Intel Tool
    from aa_intel_tool.constants import (  # pylint: disable=import-outside-toplevel
        REGEX_PATTERN,
    )

    ansiblex_destination = None
    counter = defaultdict(lambda: defaultdict(int))
    eve_ids = defaultdict(list)

    # Let's split this list up
    #
    # line.group(1) => Item ID
    # line.group(2) => Name
    # line.group(3) => Ship Class / Structure Type
    # line.group(4) => Distance
    #
    # Loop through all lines
    for entry in scan_data:
        # Apparently you can copy/paste a tab into the ship name, which will cause the split by tab to fail.
        # The regex is detecting the D-Scan correctly though. But splitting by tab might put the ship class as distance.
        # See https://github.com/ppfeufer/aa-intel-tool/issues/82
        #
        # This is why we use re.search() to get the parts of the D-Scan entry, instead of re-split()
        #
        # Thanks CCP for sanitizing your inputs! 😂
        line = re.search(pattern=REGEX_PATTERN["dscan"], string=entry)
        entry_id = int(line.group(1))

        counter["all"][entry_id] += 1

        # Check if the entry is "on grid" or not
        if _is_on_grid(distance=line.group(4)):
            counter["ongrid"][entry_id] += 1

            # If it is an Ansiblex Jump Gate, get its destination system
            if entry_id == UpwellStructureId.ANSIBLEX_JUMP_GATE:
                ansiblex_destination = _get_ansiblex_jumpgate_destination(
                    ansiblex_name=line.group(2)
                )

            eve_ids["ongrid"].append(entry_id)
        else:
            counter["offgrid"][entry_id] += 1
            eve_ids["offgrid"].append(entry_id)

        eve_ids["all"].append(entry_id)

    return ansiblex_destination, counter, eve_ids


def parse(scan_data: list) -> Scan:
    """
    Parse D-Scan

    :param scan_data:
    :type scan_data:
    :return:
    :rtype:
    """

    # Only parse the d-scan if the module is enabled
    if not AppSettings.INTELTOOL_ENABLE_MODULE_DSCAN:
        raise ParserError(message=str(_("The D-Scan module is currently disabled.")))

    parsed_data = {}
    ansiblex_destination, counter, eve_ids = _get_scan_details(scan_data=scan_data)

    eve_types = EveType.objects.bulk_get_or_create_esi(
        ids=set(eve_ids["all"]), include_children=True
    ).values_list("id", "name", "eve_group__id", "eve_group__name", "mass", named=True)

    # Parse the data parts
    ships = _get_ships(eve_types=eve_types, counter=counter)
    upwell_structures = _get_upwell_structures_on_grid(
        eve_types=eve_types,
        counter=counter,
        ansiblex_destination=ansiblex_destination,
    )
    deployables = _get_deployables_on_grid(eve_types=eve_types, counter=counter)
    starbases = _get_starbases_on_grid(eve_types=eve_types, counter=counter)

    # Add parsed data when available
    sections = {
        "all": (ScanData.Section.SHIPLIST, ships["all"]),
        "ongrid": (ScanData.Section.SHIPLIST_ON_GRID, ships["ongrid"]),
        "offgrid": (ScanData.Section.SHIPLIST_OFF_GRID, ships["offgrid"]),
        "shiptypes": (ScanData.Section.SHIPTYPES, ships["types"]),
        "structures_on_grid": (ScanData.Section.STRUCTURES_ON_GRID, upwell_structures),
        "deployables": (ScanData.Section.DEPLOYABLES_ON_GRID, deployables),
        "starbases": (ScanData.Section.STARBASES_ON_GRID, starbases),
    }

    for key, (section, data) in sections.items():
        if data:
            parsed_data[key] = {"section": section, "data": data}

    return safe_scan_to_db(scan_type=Scan.Type.DSCAN, parsed_data=parsed_data)



"""
DB helper for our parser
"""

# AA Intel Tool
from aa_intel_tool.models import Scan, ScanData


def safe_scan_to_db(scan_type: Scan.Type, parsed_data: dict) -> Scan:
    """
    Saving scan data to the DB

    :param scan_type:
    :type scan_type:
    :param parsed_data:
    :type parsed_data:
    :return:
    :rtype:
    """

    # Creating a new Scan object and saving it
    new_scan = Scan.objects.create(scan_type=scan_type)

    # Creating and saving the associated ScanData objects
    ScanData.objects.bulk_create(
        [
            ScanData(
                scan=new_scan,
                section=scan_data["section"],
                processed_data=scan_data["data"],
            )
            for scan_data in parsed_data.values()
        ]
    )

    # Return the Scan object
    return new_scan



"""
Tests for our models
"""

# Standard Library
from unittest.mock import patch

# Django
from django.db import IntegrityError
from django.test import TestCase

# AA Intel Tool
from aa_intel_tool.models import Scan, ScanData


class TestScanModel(TestCase):
    """
    Test the Scan model
    """

    def test_generates_unique_hash(self):
        """
        Test that the hash is unique

        :return:
        :rtype:
        """

        hash1 = Scan.generate_scan_hash()
        hash2 = Scan.generate_scan_hash()

        self.assertNotEqual(hash1, hash2)

    def test_generates_hash_of_correct_length(self):
        """
        Test that the hash is of the correct length

        :return:
        :rtype:
        """

        scan_hash = Scan.generate_scan_hash()

        self.assertEqual(len(scan_hash), 30)

    def test_generates_hash_not_in_database(self):
        """
        Test that the generated hash is not in the database

        :return:
        :rtype:
        """

        existing_scan = Scan.objects.create(raw_data="test data")
        scan_hash = Scan.generate_scan_hash()

        self.assertNotEqual(scan_hash, existing_scan.hash)

    def test_handles_collision_by_generating_new_hash(self):
        """
        Test that the hash collision is handled by generating a new hash

        :return:
        :rtype:
        """

        with patch("aa_intel_tool.models.Scan.objects.filter") as mock_filter:
            # Simulate hash collision by returning True for the first two calls
            mock_filter.return_value.exists.side_effect = [True, True, False]

            new_hash = Scan.generate_scan_hash()

            self.assertIsNotNone(new_hash)
            self.assertEqual(len(new_hash), 30)
            self.assertNotIn(new_hash, ["collision_hash1", "collision_hash2"])

    def test_saves_with_generated_hash(self):
        """
        Test that the hash is generated on save

        :return:
        :rtype:
        """

        scan = Scan(raw_data="test data")
        scan.save()

        self.assertIsNotNone(scan.hash)
        self.assertEqual(len(scan.hash), 30)

    def test_generates_new_hash_if_empty_on_save(self):
        """
        Test that a new hash is generated if the hash is empty on save

        :return:
        :rtype:
        """

        scan = Scan(hash="", raw_data="test data")
        scan.save()

        self.assertNotEqual(scan.hash, "")
        self.assertEqual(len(scan.hash), 30)

    def test_model_string_names(self):
        """
        Test model string name

        :return:
        :rtype:
        """

        scan = Scan(
            raw_data="Foobar",
        )
        scan.save()
        expected_hash = scan.hash

        self.assertEqual(first=str(scan), second=expected_hash)

        scan.raw_data = "BarFoo"
        scan.save()

        self.assertEqual(first=str(scan), second=expected_hash)

        scan.hash = ""
        scan.save()

        self.assertNotEqual(first=str(scan), second="")
        self.assertNotEqual(first=str(scan), second=expected_hash)


class TestScanDataModel(TestCase):
    """
    Test the ScanData model
    """

    def test_creates_scan_data_with_valid_section(self):
        """
        Test that the ScanData model can be created with a valid section

        :return:
        :rtype:
        """

        scan = Scan.objects.create(raw_data="test data")
        scan_data = ScanData.objects.create(
            scan=scan,
            section=ScanData.Section.PILOTLIST,
            processed_data={"key": "value"},
        )

        self.assertEqual(scan_data.section, ScanData.Section.PILOTLIST)
        self.assertEqual(scan_data.processed_data, {"key": "value"})

    def test_defaults_to_invalid_section(self):
        """
        Test that the ScanData model defaults to INVALID section if not specified

        :return:
        :rtype:
        """

        scan = Scan.objects.create(raw_data="test data")
        scan_data = ScanData.objects.create(scan=scan, processed_data={"key": "value"})

        self.assertEqual(scan_data.section, ScanData.Section.INVALID)

    def test_allows_null_scan(self):
        """
        Test that the ScanData model allows null scan

        :return:
        :rtype:
        """

        scan_data = ScanData.objects.create(
            section=ScanData.Section.PILOTLIST, processed_data={"key": "value"}
        )

        self.assertIsNone(scan_data.scan)

    def test_enforces_unique_together_constraint(self):
        """
        Test that the ScanData model enforces unique together constraint on scan and section and raises IntegrityError

        :return:
        :rtype:
        """

        scan = Scan.objects.create(raw_data="test data")
        ScanData.objects.create(
            scan=scan,
            section=ScanData.Section.PILOTLIST,
            processed_data={"key": "value"},
        )

        with self.assertRaises(IntegrityError):
            ScanData.objects.create(
                scan=scan,
                section=ScanData.Section.PILOTLIST,
                processed_data={"key": "another value"},
            )

    def test_allows_different_sections_for_same_scan(self):
        """
        Test that the ScanData model allows different sections for the same scan

        :return:
        :rtype:
        """

        scan = Scan.objects.create(raw_data="test data")
        scan_data1 = ScanData.objects.create(
            scan=scan,
            section=ScanData.Section.PILOTLIST,
            processed_data={"key": "value"},
        )
        scan_data2 = ScanData.objects.create(
            scan=scan,
            section=ScanData.Section.CORPORATIONLIST,
            processed_data={"key": "value"},
        )

        self.assertNotEqual(scan_data1.section, scan_data2.section)



"""
Testing the parsers
"""

# Standard Library
from unittest.mock import MagicMock, patch

# Django
from django.test import TestCase

# Alliance Auth (External Libs)
from eveuniverse.models import EveEntity

# AA Intel Tool
from aa_intel_tool.exceptions import ParserError
from aa_intel_tool.parser.general import check_intel_type, parse_intel
from aa_intel_tool.parser.module.chatlist import (
    _get_character_info,
    _get_unaffiliated_alliance_info,
    _parse_alliance_info,
    _parse_character_info,
    _parse_chatscan_data,
    _parse_corporation_info,
)
from aa_intel_tool.tests.utils import (
    load_chatscan_faulty_txt,
    load_chatscan_txt,
    load_dscan_txt,
    load_fleetcomp_txt,
)


class TestCheckIntelType(TestCase):
    """
    Test the check_intel_type function
    """

    def test_check_intel_type_dscan(self):
        """
        Test should return 'dscan' as the expected intel type

        :return:
        :rtype:
        """

        form_data = load_dscan_txt()
        scan_data = str(form_data).splitlines()

        intel_type = check_intel_type(scan_data=scan_data)
        expected_intel_type = "dscan"

        self.assertEqual(first=intel_type, second=expected_intel_type)

    def test_check_intel_type_chatlist(self):
        """
        Test should return 'chatlist' as the expected intel type

        :return:
        :rtype:
        """

        form_data = load_chatscan_txt()
        scan_data = str(form_data).splitlines()

        intel_type = check_intel_type(scan_data=scan_data)
        expected_intel_type = "chatlist"

        self.assertEqual(first=intel_type, second=expected_intel_type)

    def test_check_intel_type_fleetcomp(self):
        """
        Test should return 'fleetcomp' as the expected intel type

        :return:
        :rtype:
        """

        form_data = load_fleetcomp_txt()
        scan_data = str(form_data).splitlines()

        intel_type = check_intel_type(scan_data=scan_data)
        expected_intel_type = "fleetcomp"

        self.assertEqual(first=intel_type, second=expected_intel_type)

    def test_check_intel_type_malformed_data(self):
        """
        Test should throw a ParserError as the expected intel type
        This happens when invalid data has been posted

        :return:
        :rtype:
        """

        form_data = load_chatscan_faulty_txt()
        scan_data = str(form_data).splitlines()

        expected_exception = ParserError
        expected_message = "A parser error occurred » No suitable parser found. Input is not a supported intel type or malformed …"  # pylint: disable=line-too-long

        with self.assertRaises(expected_exception=expected_exception):
            check_intel_type(scan_data=scan_data)

        with self.assertRaisesMessage(
            expected_exception=expected_exception, expected_message=expected_message
        ):
            check_intel_type(scan_data=scan_data)

    def test_parse_intel_with_invalid_form_data(self):
        """
        Test should return a ParserError as parsed intel data for invalid form data

        :return:
        :rtype:
        """

        form_data = load_chatscan_faulty_txt()

        expected_exception = ParserError
        expected_message = "A parser error occurred » No suitable parser found. Input is not a supported intel type or malformed …"  # pylint: disable=line-too-long

        with self.assertRaises(ParserError):
            parse_intel(form_data=form_data)

        with self.assertRaisesMessage(
            expected_exception=expected_exception, expected_message=expected_message
        ):
            parse_intel(form_data=form_data)

    def test_parse_intel_empty_form_data(self):
        """
        Test should throw a ParserError as parsed intel data for empty form data

        :return:
        :rtype:
        """

        form_data = ""

        expected_exception = ParserError
        expected_message = "A parser error occurred » No data to parse …"

        with self.assertRaises(expected_exception=expected_exception):
            parse_intel(form_data=form_data)

        with self.assertRaisesMessage(
            expected_exception=expected_exception, expected_message=expected_message
        ):
            parse_intel(form_data=form_data)


class TestParseIntel(TestCase):
    """
    Test the parse_intel function
    """

    @patch("aa_intel_tool.parser.general.check_intel_type")
    @patch(
        "aa_intel_tool.parser.general.SUPPORTED_INTEL_TYPES",
        {"dscan": {"parser": MagicMock(return_value=MagicMock(hash="hash1"))}},
    )
    def test_parses_valid_dscan_data(self, mock_check_intel_type):
        """
        Test should return the hash of the parsed intel data for valid dscan data

        :param mock_check_intel_type:
        :type mock_check_intel_type:
        :return:
        :rtype:
        """

        mock_check_intel_type.return_value = "dscan"
        form_data = load_dscan_txt()
        result = parse_intel(form_data)

        self.assertEqual(result, "hash1")

    @patch("aa_intel_tool.parser.general.check_intel_type")
    @patch(
        "aa_intel_tool.parser.general.SUPPORTED_INTEL_TYPES",
        {"fleetcomp": {"parser": MagicMock(return_value=MagicMock(hash="hash2"))}},
    )
    def test_parses_valid_fleetcomp_data(self, mock_check_intel_type):
        """
        Test should return the hash of the parsed intel data for valid fleetcomp data

        :param mock_check_intel_type:
        :type mock_check_intel_type:
        :return:
        :rtype:
        """

        mock_check_intel_type.return_value = "fleetcomp"
        form_data = load_fleetcomp_txt()
        result = parse_intel(form_data)

        self.assertEqual(result, "hash2")

    @patch("aa_intel_tool.parser.general.check_intel_type")
    @patch(
        "aa_intel_tool.parser.general.SUPPORTED_INTEL_TYPES",
        {"chatscan": {"parser": MagicMock(return_value=MagicMock(hash="hash3"))}},
    )
    def test_parses_valid_chatscan_data(self, mock_check_intel_type):
        """
        Test should return the hash of the parsed intel data for valid chatscan data

        :param mock_check_intel_type:
        :type mock_check_intel_type:
        :return:
        :rtype:
        """

        mock_check_intel_type.return_value = "chatscan"
        form_data = load_chatscan_txt()
        result = parse_intel(form_data)

        self.assertEqual(result, "hash3")

    @patch("aa_intel_tool.parser.general.check_intel_type")
    def test_raises_error_for_invalid_data(self, mock_check_intel_type):
        """
        Test should throw a ParserError as parsed intel data for invalid data

        :param mock_check_intel_type:
        :type mock_check_intel_type:
        :return:
        :rtype:
        """

        mock_check_intel_type.side_effect = ParserError("Invalid data")
        form_data = load_chatscan_faulty_txt()

        with self.assertRaises(ParserError):
            parse_intel(form_data)

    def test_raises_error_for_empty_data(self):
        """
        Test should throw a ParserError as parsed intel data for empty form data

        :return:
        :rtype:
        """

        form_data = ""

        with self.assertRaises(ParserError):
            parse_intel(form_data)


class TestParseChatScanData(TestCase):
    """
    Test the _parse_chatscan_data function
    """

    @patch("aa_intel_tool.parser.module.chatlist._parse_alliance_info")
    @patch("aa_intel_tool.parser.module.chatlist._parse_corporation_info")
    @patch("aa_intel_tool.parser.module.chatlist._parse_character_info")
    def test_parses_valid_characters(
        self,
        mock_parse_character_info,
        mock_parse_corporation_info,
        mock_parse_alliance_info,
    ):
        """
        Test should parse valid chatscan data and return a dictionary

        :param mock_parse_character_info:
        :type mock_parse_character_info:
        :param mock_parse_corporation_info:
        :type mock_parse_corporation_info:
        :param mock_parse_alliance_info:
        :type mock_parse_alliance_info:
        :return:
        :rtype:
        """

        eve_characters = MagicMock()
        eve_characters.__iter__.return_value = [
            MagicMock(
                character_name="Character1",
                alliance_name="Alliance1",
                corporation_name="Corp1",
            ),
            MagicMock(
                character_name="Character2",
                alliance_name="Alliance2",
                corporation_name="Corp2",
            ),
        ]
        mock_parse_character_info.side_effect = lambda eve_character: {
            "name": eve_character.character_name
        }
        mock_parse_corporation_info.side_effect = lambda eve_character: {
            "name": eve_character.corporation_name
        }
        mock_parse_alliance_info.side_effect = lambda eve_character: {
            "name": eve_character.alliance_name
        }

        result = _parse_chatscan_data(eve_characters)

        self.assertEqual(len(result["pilots"]), 2)
        self.assertEqual(len(result["corporations"]), 2)
        self.assertEqual(len(result["alliances"]), 2)

    @patch("aa_intel_tool.parser.module.chatlist._parse_alliance_info")
    @patch("aa_intel_tool.parser.module.chatlist._parse_corporation_info")
    @patch("aa_intel_tool.parser.module.chatlist._parse_character_info")
    def test_handles_unaffiliated_characters(
        self,
        mock_parse_character_info,
        mock_parse_corporation_info,
        mock_parse_alliance_info,
    ):
        """
        Test should handle unaffiliated characters and return "Unaffiliated" as the alliance name

        :param mock_parse_character_info:
        :type mock_parse_character_info:
        :param mock_parse_corporation_info:
        :type mock_parse_corporation_info:
        :param mock_parse_alliance_info:
        :type mock_parse_alliance_info:
        :return:
        :rtype:
        """

        eve_characters = MagicMock()
        eve_characters.__iter__.return_value = [
            MagicMock(
                character_name="Character1",
                alliance_name=None,
                corporation_name="Corp1",
            )
        ]
        mock_parse_character_info.side_effect = lambda eve_character: {
            "name": eve_character.character_name
        }
        mock_parse_corporation_info.side_effect = lambda eve_character: {
            "name": eve_character.corporation_name
        }
        mock_parse_alliance_info.side_effect = lambda eve_character: {
            "name": "Unaffiliated"
        }

        result = _parse_chatscan_data(eve_characters)

        self.assertEqual(result["alliances"][0]["name"], "Unaffiliated")

    @patch("aa_intel_tool.parser.module.chatlist._parse_alliance_info")
    @patch("aa_intel_tool.parser.module.chatlist._parse_corporation_info")
    @patch("aa_intel_tool.parser.module.chatlist._parse_character_info")
    def test_counts_characters_correctly(
        self,
        mock_parse_character_info,
        mock_parse_corporation_info,
        mock_parse_alliance_info,
    ):
        """
        Test should count characters correctly and return the correct counts for alliances and corporations

        :param mock_parse_character_info:
        :type mock_parse_character_info:
        :param mock_parse_corporation_info:
        :type mock_parse_corporation_info:
        :param mock_parse_alliance_info:
        :type mock_parse_alliance_info:
        :return:
        :rtype:
        """

        eve_characters = MagicMock()
        eve_characters.__iter__.return_value = [
            MagicMock(
                character_name="Character1",
                alliance_name="Alliance1",
                corporation_name="Corp1",
            ),
            MagicMock(
                character_name="Character2",
                alliance_name="Alliance1",
                corporation_name="Corp1",
            ),
            MagicMock(
                character_name="Character3",
                alliance_name="Alliance2",
                corporation_name="Corp2",
            ),
        ]
        mock_parse_character_info.side_effect = lambda eve_character: {
            "name": eve_character.character_name
        }
        mock_parse_corporation_info.side_effect = lambda eve_character: {
            "name": eve_character.corporation_name
        }
        mock_parse_alliance_info.side_effect = lambda eve_character: {
            "name": eve_character.alliance_name
        }

        result = _parse_chatscan_data(eve_characters)

        self.assertEqual(result["alliances"][0]["count"], 2)
        self.assertEqual(result["corporations"][0]["count"], 2)
        self.assertEqual(result["alliances"][1]["count"], 1)
        self.assertEqual(result["corporations"][1]["count"], 1)


class TestParseCharacterInfo(TestCase):
    """
    Test the _parse_character_info function
    """

    @patch("aa_intel_tool.parser.module.chatlist._parse_alliance_info")
    @patch("aa_intel_tool.parser.module.chatlist._parse_corporation_info")
    def test_parses_valid_character_info(
        self, mock_parse_corporation_info, mock_parse_alliance_info
    ):
        """
        Test should parse valid character info and return a dictionary

        :param mock_parse_corporation_info:
        :type mock_parse_corporation_info:
        :param mock_parse_alliance_info:
        :type mock_parse_alliance_info:
        :return:
        :rtype:
        """

        eve_character = MagicMock(
            character_id=123,
            character_name="Character1",
            portrait_url_32="portrait_url",
        )
        mock_parse_corporation_info.return_value = {"name": "Corp1"}
        mock_parse_alliance_info.return_value = {"name": "Alliance1"}

        result = _parse_character_info(eve_character)

        self.assertEqual(result["id"], 123)
        self.assertEqual(result["name"], "Character1")
        self.assertEqual(result["portrait"], "portrait_url")
        self.assertEqual(result["corporation"]["name"], "Corp1")
        self.assertEqual(result["alliance"]["name"], "Alliance1")

    @patch("aa_intel_tool.parser.module.chatlist._parse_alliance_info")
    @patch("aa_intel_tool.parser.module.chatlist._parse_corporation_info")
    def test_handles_missing_alliance_info(
        self, mock_parse_corporation_info, mock_parse_alliance_info
    ):
        """
        Test should handle missing alliance info and return "Unaffiliated" as the alliance name

        :param mock_parse_corporation_info:
        :type mock_parse_corporation_info:
        :param mock_parse_alliance_info:
        :type mock_parse_alliance_info:
        :return:
        :rtype:
        """

        eve_character = MagicMock(
            character_id=123,
            character_name="Character1",
            portrait_url_32="portrait_url",
        )
        mock_parse_corporation_info.return_value = {"name": "Corp1"}
        mock_parse_alliance_info.return_value = {"name": "Unaffiliated"}

        result = _parse_character_info(eve_character)

        self.assertEqual(result["alliance"]["name"], "Unaffiliated")


class TestParseCorporationInfo(TestCase):
    """
    Test the _parse_corporation_info function
    """

    @patch("aa_intel_tool.parser.module.chatlist._parse_alliance_info")
    def test_parses_valid_corporation_info(self, mock_parse_alliance_info):
        """
        Test should parse valid corporation info and return a dictionary

        :param mock_parse_alliance_info:
        :type mock_parse_alliance_info:
        :return:
        :rtype:
        """

        eve_character = MagicMock(
            corporation_id=456,
            corporation_name="Corp1",
            corporation_ticker="C1",
            corporation_logo_url_32="corp_logo_url",
        )
        mock_parse_alliance_info.return_value = {"name": "Alliance1"}

        result = _parse_corporation_info(eve_character)

        self.assertEqual(result["id"], 456)
        self.assertEqual(result["name"], "Corp1")
        self.assertEqual(result["ticker"], "C1")
        self.assertEqual(result["logo"], "corp_logo_url")
        self.assertEqual(result["alliance"]["name"], "Alliance1")

    @patch("aa_intel_tool.parser.module.chatlist._parse_alliance_info")
    def test_handles_missing_alliance_info(self, mock_parse_alliance_info):
        """
        Test should handle missing alliance info and return "Unaffiliated" as the alliance name

        :param mock_parse_alliance_info:
        :type mock_parse_alliance_info:
        :return:
        :rtype:
        """

        eve_character = MagicMock(
            corporation_id=456,
            corporation_name="Corp1",
            corporation_ticker="C1",
            corporation_logo_url_32="corp_logo_url",
        )
        mock_parse_alliance_info.return_value = {"name": "Unaffiliated"}

        result = _parse_corporation_info(eve_character, with_alliance_info=False)

        self.assertNotIn("alliance", result)


class TestParseAllianceInfo(TestCase):
    """
    Test the _parse_alliance_info function
    """

    def test_parses_valid_alliance_info(self):
        """
        Test should parse valid alliance info and return a dictionary

        :return:
        :rtype:
        """

        eve_character = MagicMock(
            alliance_id=789,
            alliance_name="Alliance1",
            alliance_ticker="A1",
            alliance_logo_url_32="alliance_logo_url",
        )

        result = _parse_alliance_info(eve_character)

        self.assertEqual(result["id"], 789)
        self.assertEqual(result["name"], "Alliance1")
        self.assertEqual(result["ticker"], "A1")
        self.assertEqual(result["logo"], "alliance_logo_url")
        self.assertIn("dotlan", result)
        self.assertIn("zkillboard", result)

    def test_handles_unaffiliated_alliance(self):
        """
        Test should handle unaffiliated alliance info and return "Unaffiliated" as the alliance name

        :return:
        :rtype:
        """

        eve_character = MagicMock(alliance_id=None)

        result = _parse_alliance_info(eve_character)

        self.assertEqual(result["id"], 1)
        self.assertEqual(result["name"], "")
        self.assertEqual(result["ticker"], "")
        self.assertIn("logo", result)

    @patch("aa_intel_tool.parser.module.chatlist.dotlan.alliance_url")
    @patch("aa_intel_tool.parser.module.chatlist.zkillboard.alliance_url")
    def test_includes_evelinks(self, mock_zkillboard, mock_dotlan):
        """
        Test should include zkillboard and dotlan links in the result

        :param mock_zkillboard:
        :type mock_zkillboard:
        :param mock_dotlan:
        :type mock_dotlan:
        :return:
        :rtype:
        """

        eve_character = MagicMock(
            alliance_id=789,
            alliance_name="Alliance1",
            alliance_ticker="A1",
            alliance_logo_url_32="alliance_logo_url",
        )
        mock_dotlan.return_value = "dotlan_url"
        mock_zkillboard.return_value = "zkillboard_url"

        result = _parse_alliance_info(eve_character, with_evelinks=True)

        self.assertEqual(result["dotlan"], "dotlan_url")
        self.assertEqual(result["zkillboard"], "zkillboard_url")

    def test_excludes_evelinks(self):
        """
        Test should exclude zkillboard and dotlan links from the result

        :return:
        :rtype:
        """

        eve_character = MagicMock(
            alliance_id=789,
            alliance_name="Alliance1",
            alliance_ticker="A1",
            alliance_logo_url_32="alliance_logo_url",
        )

        result = _parse_alliance_info(eve_character, with_evelinks=False)

        self.assertNotIn("dotlan", result)
        self.assertNotIn("zkillboard", result)


class TestGetUnaffiliatedAllianceInfo(TestCase):
    """
    Test the _get_unaffiliated_alliance_info function
    """

    @patch("aa_intel_tool.parser.module.chatlist.eveimageserver.alliance_logo_url")
    def test_returns_correct_alliance_info(self, mock_alliance_logo_url):
        """
        Test should return the correct information for the unaffiliated alliance

        :param mock_alliance_logo_url:
        :type mock_alliance_logo_url:
        :return:
        :rtype:
        """

        mock_alliance_logo_url.return_value = "mock_logo_url"

        result = _get_unaffiliated_alliance_info()

        self.assertEqual(result["id"], 1)
        self.assertEqual(result["name"], "")
        self.assertEqual(result["ticker"], "")
        self.assertEqual(result["logo"], "mock_logo_url")

    @patch("aa_intel_tool.parser.module.chatlist.eveimageserver.alliance_logo_url")
    def handles_logo_url_generation(self, mock_alliance_logo_url):
        """
        Test should generate the correct logo URL for the unaffiliated alliance

        :param mock_alliance_logo_url:
        :type mock_alliance_logo_url:
        :return:
        :rtype:
        """

        mock_alliance_logo_url.return_value = "mock_logo_url"

        result = _get_unaffiliated_alliance_info()

        mock_alliance_logo_url.assert_called_once_with(alliance_id=1, size=32)
        self.assertEqual(result["logo"], "mock_logo_url")


class TestGetCharacterInfo(TestCase):
    """
    Test the _get_character_info function
    """

    @patch("aa_intel_tool.parser.module.chatlist.EveCharacter.objects.filter")
    @patch("aa_intel_tool.parser.module.chatlist.get_or_create_character")
    def test_returns_existing_characters(self, mock_get_or_create, mock_filter):
        """
        Test should return existing characters from the database

        :param mock_get_or_create:
        :type mock_get_or_create:
        :param mock_filter:
        :type mock_filter:
        :return:
        :rtype:
        """

        # Mock the QuerySet returned by filter
        mock_queryset = MagicMock()
        mock_queryset.exclude.return_value = mock_queryset
        mock_queryset.count.return_value = 0
        mock_queryset.__iter__.return_value = iter([])
        mock_filter.return_value = mock_queryset

        # Mock the creation of a new character
        mock_get_or_create.return_value = [MagicMock(character_name="Character1")]

        result = _get_character_info(["Character1", "Character2"])

        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].character_name, "Character1")

    @patch("aa_intel_tool.parser.module.chatlist.EveCharacter.objects.filter")
    def test_fetches_characters_from_eveuniverse(self, mock_filter):
        """
        Test should fetch characters from the Eve Universe API

        :param mock_filter:
        :type mock_filter:
        :return:
        :rtype:
        """

        mock_filter.return_value.exclude.return_value.count.return_value = 0

        with patch(
            "aa_intel_tool.parser.module.chatlist.EveEntity.objects.fetch_by_names_esi"
        ) as mock_fetch:
            mock_fetch.return_value.filter.return_value.values_list.return_value = [1]
            with patch(
                "aa_intel_tool.parser.module.chatlist.get_or_create_character"
            ) as mock_get_or_create:
                mock_get_or_create.return_value = [
                    MagicMock(character_name="Character1")
                ]

                result = _get_character_info(["Character1"])

                self.assertEqual(len(result), 1)
                self.assertEqual(result[0].character_name, "Character1")

    @patch("aa_intel_tool.parser.module.chatlist.EveCharacter.objects.filter")
    def test_handles_no_characters_found(self, mock_filter):
        """
        Test should raise a ParserError when no characters are found

        :param mock_filter:
        :type mock_filter:
        :return:
        :rtype:
        """

        mock_filter.return_value.exclude.return_value.count.return_value = 0

        with patch(
            "aa_intel_tool.parser.module.chatlist.EveEntity.objects.fetch_by_names_esi"
        ) as mock_fetch:
            mock_fetch.return_value.filter.return_value.values_list.return_value = []
            with self.assertRaises(ParserError):
                _get_character_info(["UnknownCharacter"])

    @patch("aa_intel_tool.parser.module.chatlist.EveCharacter.objects.filter")
    def test_handles_eveuniverse_fetch_error(self, mock_filter):
        """
        Test should raise a ParserError when there is an error fetching from the Eve Universe API

        :param mock_filter:
        :type mock_filter:
        :return:
        :rtype:
        """

        mock_filter.return_value.exclude.return_value.count.return_value = 0

        with patch(
            "aa_intel_tool.parser.module.chatlist.EveEntity.objects.fetch_by_names_esi"
        ) as mock_fetch:
            mock_fetch.side_effect = EveEntity.DoesNotExist
            with self.assertRaises(ParserError):
                _get_character_info(["Character1"])



"""
Little helper functions to deal with Eve characters
"""

# Standard Library
from collections.abc import Iterable

# Django
from django.db.models import QuerySet

# Alliance Auth
from allianceauth.eveonline.models import (
    EveAllianceInfo,
    EveCharacter,
    EveCorporationInfo,
)
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Intel Tool
from aa_intel_tool import __title__

logger = LoggerAddTag(my_logger=get_extension_logger(name=__name__), prefix=__title__)


def _create_alliance(alliance_ids: Iterable[int]) -> None:
    """
    Bulk creation of EveAllianceInfo objects

    :param alliance_ids:
    :type alliance_ids:
    :return:
    :rtype:
    """

    alliance_ids = set(alliance_ids)
    existing_alliance_ids = set(
        EveAllianceInfo.objects.filter(alliance_id__in=alliance_ids).values_list(
            "alliance_id", flat=True
        )
    )

    alliances_to_fetch = alliance_ids - existing_alliance_ids

    if alliances_to_fetch:
        logger.debug(
            f"{len(alliances_to_fetch)} EveAllianceInfo object(s) need to be created …"
        )

        for loop_count, alliance_id in enumerate(alliances_to_fetch, start=1):
            alliance = EveAllianceInfo.objects.create_alliance(alliance_id=alliance_id)
            logger.debug(
                f"({loop_count}/{len(alliances_to_fetch)}) "
                f"EveAllianceInfo object created for: {alliance.alliance_name}"
            )


def _create_corporation(corporation_ids: Iterable[int]) -> None:
    """
    Bulk creation of EveCorporationInfo objects

    :param corporation_ids:
    :type corporation_ids:
    :return:
    :rtype:
    """

    corporation_ids = set(corporation_ids)
    existing_corporation_ids = set(
        EveCorporationInfo.objects.filter(
            corporation_id__in=corporation_ids
        ).values_list("corporation_id", flat=True)
    )

    corporations_to_fetch = corporation_ids - existing_corporation_ids

    if corporations_to_fetch:
        logger.debug(
            f"{len(corporations_to_fetch)} EveCorporationInfo object(s) need to be created …"
        )

        for loop_count, corporation_id in enumerate(corporations_to_fetch, start=1):
            corporation = EveCorporationInfo.objects.create_corporation(
                corp_id=corporation_id
            )

            logger.debug(
                f"({loop_count}/{len(corporations_to_fetch)}) "
                f"EveCorporationInfo object created for: {corporation.corporation_name}"
            )


def _create_character(
    character_ids: Iterable[int], with_affiliation: bool = True
) -> None:
    """
    Bulk creation of EveCharacter objects

    :param character_ids:
    :type character_ids:
    :param with_affiliation:
    :type with_affiliation:
    :return:
    :rtype:
    """

    character_ids = set(character_ids)
    tmp_affiliation_ids = {"alliance": set(), "corporation": set()}

    logger.info(f"{len(character_ids)} EveCharacter object(s) need to be created …")

    for loop_count, character_id in enumerate(character_ids, start=1):
        # Create character
        character = EveCharacter.objects.create_character(character_id=character_id)

        logger.debug(
            f"({loop_count}/{len(character_ids)}) "
            f"EveCharacter object created for: {character.character_name}"
        )

        affiliation_key = "alliance" if character.alliance_id else "corporation"
        tmp_affiliation_ids[affiliation_key].add(
            character.alliance_id or character.corporation_id
        )

    if with_affiliation:
        if tmp_affiliation_ids["alliance"]:
            _create_alliance(tmp_affiliation_ids["alliance"])

        if tmp_affiliation_ids["corporation"]:
            _create_corporation(tmp_affiliation_ids["corporation"])


def get_or_create_character(
    character_ids: Iterable[int] = None,
) -> QuerySet[EveCharacter]:
    """
    This function takes a list of character IDs and checks if the characters already
    exist in Auth and creates them with their corporation and alliance associations,
    if needed.

    :param character_ids:
    :type character_ids:
    :return:
    :rtype:
    """

    character_ids = set(character_ids or [])

    logger.debug(
        msg=f"Getting information for {len(character_ids)} character(s) from AA …"
    )

    existing_character_ids = set(
        EveCharacter.objects.filter(character_id__in=character_ids).values_list(
            "character_id", flat=True
        )
    )

    character_ids_to_fetch = character_ids - existing_character_ids

    if character_ids_to_fetch:
        _create_character(character_ids=character_ids_to_fetch, with_affiliation=True)

    return EveCharacter.objects.filter(character_id__in=character_ids)



"""
Our forms
"""

# Django
from django import forms
from django.utils.translation import gettext_lazy as _


class IntelForm(forms.Form):
    """
    Intel form
    """

    eve_intel = forms.CharField(
        widget=forms.Textarea(
            attrs={
                "rows": 15,
                "input_type": "textarea",
                "placeholder": _("Paste here …"),
                "autofocus": "autofocus",
            }
        ),
        required=True,
        label="",  # Make sure there is no form label, we don't need it
    )



"""
General parser functions
"""

# Standard Library
import re

# Django
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Intel Tool
from aa_intel_tool import __title__
from aa_intel_tool.constants import SUPPORTED_INTEL_TYPES
from aa_intel_tool.exceptions import ParserError

logger = LoggerAddTag(my_logger=get_extension_logger(name=__name__), prefix=__title__)


def check_intel_type(scan_data: list) -> str:
    """
    Check which intel type we have

    :param scan_data:
    :type scan_data:
    :return:
    :rtype:
    """

    for intel_type, intel_type_attributes in SUPPORTED_INTEL_TYPES.items():
        if all(
            re.match(pattern=intel_type_attributes["pattern"], string=string)
            for string in scan_data
        ):
            logger.debug(msg=f"Detected intel type: {intel_type_attributes['name']}")

            return intel_type

    raise ParserError(
        message=str(
            _(
                "No suitable parser found. Input is not a supported intel type or malformed …"
            )
        )
    )


def parse_intel(form_data: str) -> str:
    """
    Parse intel

    :param form_data:
    :type form_data:
    :return:
    :rtype:
    """
    scan_data = form_data.splitlines()

    if not scan_data:
        raise ParserError(message=str(_("No data to parse …")))

    try:
        intel_type = check_intel_type(scan_data=scan_data)
        new_scan = SUPPORTED_INTEL_TYPES[intel_type]["parser"](scan_data=scan_data)
    except ParserError as exc:
        # Re-raise the Exception
        raise ParserError(message=exc.message) from exc

    new_scan.raw_data = form_data
    new_scan.save()

    return new_scan.hash



"""
Preloads data required for AA Intel Tool from ESI
"""

# Standard Library
import logging

# Django
from django.core.management import call_command
from django.core.management.base import BaseCommand

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag
from eveuniverse.constants import EveCategoryId

# AA Intel Tool
from aa_intel_tool import __title__
from aa_intel_tool.app_settings import AdditionalEveCategoryId

logger = LoggerAddTag(logging.getLogger(__name__), __title__)


class Command(BaseCommand):
    """
    Pre-loading required data
    """

    help = "Preloads data required for AA Intel Tool from ESI"

    def add_arguments(self, parser):
        """
        Adding arguments to the command parser

        :param parser:
        :type parser:
        :return:
        :rtype:
        """

        parser.add_argument(
            "--noinput",
            "--no-input",
            action="store_true",
            help="Do NOT prompt the user for input of any kind.",
        )

    def handle(self, *args, **options):  # pylint: disable=unused-argument
        """
        Start the eve type import

        :param args:
        :type args:
        :param options:
        :type options:
        :return:
        :rtype:
        """

        params = [
            "eveuniverse_load_types",
            __title__,
            "--category_id",
            str(EveCategoryId.SHIP.value),
            "--category_id",
            str(EveCategoryId.STRUCTURE.value),
            "--category_id",
            str(AdditionalEveCategoryId.DEPLOYABLE.value),
            "--category_id",
            str(AdditionalEveCategoryId.STARBASE.value),
        ]

        if options["noinput"]:
            params.append("--noinput")

        call_command(*params)
